"""
Mixin classes for Motor Devices
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~EpicsMotorDialMixin
   ~EpicsMotorEnableMixin
   ~EpicsMotorLimitsMixin
   ~EpicsMotorRawMixin
   ~EpicsMotorResolutionMixin
   ~EpicsMotorServoMixin
"""

from bluesky import plan_stubs as bps
from ophyd import Component
from ophyd import EpicsSignal
from .mixin_base import DeviceMixinBase


class EpicsMotorDialMixin(DeviceMixinBase):
    """
    add motor record's dial coordinate fields to Device

    .. index:: Ophyd Device Mixin; EpicsMotorDialMixin

    EXAMPLE::

        from ophyd import EpicsMotor
        from apstools.devices import EpicsMotorDialMixin

        class myEpicsMotor(EpicsMotorDialMixin, EpicsMotor): pass
        m1 = myEpicsMotor('xxx:m1', name='m1')
        print(m1.dial.read())

    """

    dial = Component(EpicsSignal, ".DRBV", write_pv=".DVAL")


class EpicsMotorEnableMixin(DeviceMixinBase):
    """
    mixin providing access to motor enable/disable

    .. index:: Ophyd Device Mixin; EpicsMotorEnableMixin

    EXAMPLE::

        from ophyd import EpicsMotor
        from apstools.devices import EpicsMotorEnableMixin

        class MyEpicsMotor(EpicsMotorEnableMixin, EpicsMotor): ...

        m1 = MyEpicsMotor('xxx:m1', name='m1')
        print(m1.enabled)

    In a bluesky plan::

        yield from bps.mv(m1.enable_disable, m1.MOTOR_DISABLE)
        # ... other activities
        yield from bps.mv(m1.enable_disable, m1.MOTOR_ENABLE)

    """

    enable_disable = Component(EpicsSignal, "_able", kind="omitted")

    # constants for internal use
    MOTOR_ENABLE = 0
    MOTOR_DISABLE = 1

    @property
    def enabled(self):
        return self.enable_disable.get() in (self.MOTOR_ENABLE, "Enabled")

    def enable_motor(self):
        """BLOCKING call to enable motor axis"""
        self.enable_disable.put(self.MOTOR_ENABLE)

    def disable_motor(self):
        """BLOCKING call to disable motor axis"""
        self.enable_disable.put(self.MOTOR_DISABLE)


class EpicsMotorLimitsMixin(DeviceMixinBase):
    """
    add motor record HLM & LLM fields & compatibility get_lim() and set_lim()

    .. index:: Ophyd Device Mixin; EpicsMotorLimitsMixin

    EXAMPLE::

        from ophyd import EpicsMotor
        from apstools.devices import EpicsMotorLimitsMixin

        class myEpicsMotor(EpicsMotorLimitsMixin, EpicsMotor): pass
        m1 = myEpicsMotor('xxx:m1', name='m1')
        lo = m1.get_lim(-1)
        hi = m1.get_lim(1)
        m1.set_lim(-25, -5)
        print(m1.get_lim(-1), m1.get_lim(1))
        m1.set_lim(lo, hi)
    """

    soft_limit_lo = Component(EpicsSignal, ".LLM", kind="omitted")
    soft_limit_hi = Component(EpicsSignal, ".HLM", kind="omitted")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def cb_limit_changed(value, old_value, **kwargs):
            """
            update EpicsSignal object when a limit CA monitor received
            """
            if self.connected and old_value is not None and value != old_value:
                self.user_setpoint._metadata_changed(
                    self.user_setpoint.pvname,
                    self.user_setpoint._read_pv.get_ctrlvars(),
                    from_monitor=True,
                    update=True,
                )

        self.soft_limit_lo.subscribe(cb_limit_changed)
        self.soft_limit_hi.subscribe(cb_limit_changed)

    def get_lim(self, flag):
        """
        Returns the user limit of motor

        * flag > 0: returns high limit
        * flag < 0: returns low limit
        * flag == 0: returns None

        Similar with SPEC command
        """
        if flag > 0:
            return self.soft_limit_hi.get()
        elif flag < 0:
            return self.soft_limit_lo.get()

    def set_lim(self, low, high):
        """
        Sets the low and high limits of motor

        * No action taken if motor is moving.
        * Low limit is set to lesser of (low, high)
        * High limit is set to greater of (low, high)

        Similar with SPEC command
        """
        if not self.moving:
            # update EPICS
            # fmt: off
            yield from bps.mv(
                self.soft_limit_lo, min(low, high),
                self.soft_limit_hi, max(low, high),
            )
            # fmt: on
            # ophyd metadata dictionary will update via CA monitor


class EpicsMotorServoMixin(DeviceMixinBase):
    """
    add motor record's servo loop controls to Device

    .. index:: Ophyd Device Mixin; EpicsMotorServoMixin

    EXAMPLE::

        from ophyd import EpicsMotor
        from apstools.devices import EpicsMotorServoMixin

        class myEpicsMotor(EpicsMotorServoMixin, EpicsMotor): pass
        m1 = myEpicsMotor('xxx:m1', name='m1')
        print(m1.servo.read())
    """

    # values: "Enable" or "Disable"
    servo = Component(EpicsSignal, ".CNEN", string=True)


class EpicsMotorRawMixin(DeviceMixinBase):
    """
    add motor record's raw coordinate fields to Device

    .. index:: Ophyd Device Mixin; EpicsMotorRawMixin

    EXAMPLE::

        from ophyd import EpicsMotor
        from apstools.devices import EpicsMotorRawMixin

        class myEpicsMotor(EpicsMotorRawMixin, EpicsMotor): pass
        m1 = myEpicsMotor('xxx:m1', name='m1')
        print(m1.raw.read())
    """

    raw = Component(EpicsSignal, ".RRBV", write_pv=".RVAL")


class EpicsMotorResolutionMixin(DeviceMixinBase):
    """
    Add motor record's resolution fields to motor.

    .. index:: Ophyd Device Mixin; EpicsMotorResolutionMixin

    Usually, a facility will not provide such high-level
    access to calibration parameters since these are
    associated with fixed parameters of hardware.
    For simulators, it is convenient to provide access
    so that default settings (typically low-resolution)
    from the IOC can be changed as part of the device
    setup in bluesky.

    EXAMPLE::

        from ophyd import EpicsMotor
        from apstools.devices import EpicsMotorResolutionMixin

        class myEpicsMotor(EpicsMotorResolutionMixin, EpicsMotor): pass
        m1 = myEpicsMotor('xxx:m1', name='m1')
        print(f"resolution={m1.resolution.read()}")
        print(f"steps_per_rev={m1.steps_per_rev.read()}")
        print(f"units_per_rev={m1.units_per_rev.read()}")
    """

    resolution = Component(EpicsSignal, ".MRES", kind="omitted")
    steps_per_rev = Component(EpicsSignal, ".SREV", kind="omitted")
    units_per_rev = Component(EpicsSignal, ".UREV", kind="omitted")

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
nscan plan
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~CommandFileReadError
   ~command_list_as_table
   ~execute_command_list
   ~get_command_list
   ~parse_Excel_command_file
   ~parse_text_command_file
   ~register_command_handler
   ~run_command_file
   ~summarize_command_file
"""

import logging
import os
import pyRestTable

from bluesky import plan_stubs as bps

from .. import utils

logger = logging.getLogger(__name__)
logger.info(__file__)


class CommandFileReadError(IOError):
    """
    Exception when reading a command file.

    .. index:: Bluesky Exception; CommandFileReadError
    """


def command_list_as_table(commands, show_raw=False):
    """
    format a command list as a pyRestTable.Table object
    """
    tbl = pyRestTable.Table()
    tbl.addLabel("line #")
    tbl.addLabel("action")
    tbl.addLabel("parameters")
    if show_raw:  # only the developer might use this
        tbl.addLabel("raw input")
    for command in commands:
        action, args, line_number, raw_command = command
        row = [line_number, action, ", ".join(map(str, args))]
        if show_raw:
            row.append(str(raw_command))
        tbl.addRow(row)
    return tbl


def execute_command_list(filename, commands, md=None):
    """
    plan: execute the command list

    .. index:: Bluesky Plan; execute_command_list

    The command list is a tuple described below.

    * Only recognized commands will be executed.
    * Unrecognized commands will be reported as comments.

    See example implementation with APS USAXS instrument:
    https://github.com/APS-USAXS/ipython-usaxs/blob/5db882c47d935c593968f1e2144d35bec7d0181e/profile_bluesky/startup/50-plans.py#L381-L469

    PARAMETERS

    filename
        *str* :
        Name of input text file.  Can be relative or absolute path,
        such as "actions.txt", "../sample.txt", or
        "/path/to/overnight.txt".
    commands
        *[command]* :
        List of command tuples for use in ``execute_command_list()``

    where

    command
        *tuple* :
        (action, OrderedDict, line_number, raw_command)
    action
        *str* :
        names a known action to be handled
    parameters
        *list* :
        List of parameters for the action.
        The list is empty of there are no values
    line_number
        *int* :
        line number (1-based) from the input text file
    raw_command
        *str* or *[str]* :
        contents from input file, such as:
        ``SAXS 0 0 0 blank``

    SEE ALSO

    .. autosummary::

        ~execute_command_list
        ~register_command_handler
        ~run_command_file
        ~summarize_command_file
        ~parse_Excel_command_file
        ~parse_text_command_file

    *new in apstools release 1.1.7*
    """
    full_filename = os.path.abspath(filename)

    if len(commands) == 0:
        yield from bps.null()
        return

    text = f"Command file: {filename}\n"
    text += str(command_list_as_table(commands))
    print(text)

    for command in commands:
        action, args, i, raw_command = command
        print(f"file line {i}: {raw_command}")

        _md = {}
        _md["full_filename"] = full_filename
        _md["filename"] = filename
        _md["line_number"] = i
        _md["action"] = action
        # note: args is shorter than parameters, means the same thing here
        _md["parameters"] = args

        _md.update(md or {})  # overlay with user-supplied metadata

        action = action.lower()
        if action == "tune_optics":
            # example: yield from tune_optics(md=_md)
            emsg = (
                f"This is an example.  action={raw_command}."
                "  Must define your own execute_command_list() function"
            )
            logger.warn(emsg)
            yield from bps.null()

        else:
            print(f"no handling for line {i}: {raw_command}")


def get_command_list(filename):
    """
    return command list from either text or Excel file

    SEE ALSO

    .. autosummary::

        ~execute_command_list
        ~get_command_list
        ~register_command_handler
        ~run_command_file
        ~summarize_command_file
        ~parse_Excel_command_file
        ~parse_text_command_file

    *new in apstools release 1.1.7*
    """
    full_filename = os.path.abspath(filename)
    if not os.path.exists(full_filename):
        raise IOError(f"file not found: {filename}")
    try:
        commands = parse_Excel_command_file(filename)
    except (ValueError, utils.ExcelReadError):
        try:
            commands = parse_text_command_file(filename)
        except ValueError as exc:
            raise CommandFileReadError(
                f"could not read {filename} as command list file: {exc}"
            )
    return commands


class _HandlerRegistrar:
    """
    Internal use, allows redefinition of execute_command_list().
    """

    command = None


def parse_Excel_command_file(filename):
    """
    parse an Excel spreadsheet with commands, return as command list

    TEXT view of spreadsheet (Excel file line numbers shown)::

        [1] List of sample scans to be run
        [2]
        [3]
        [4] scan    sx  sy  thickness   sample name
        [5] FlyScan 0   0   0   blank
        [6] FlyScan 5   2   0   blank

    PARAMETERS

    filename
        *str* :
        Name of input Excel spreadsheet file.  Can be relative or absolute path,
        such as "actions.xslx", "../sample.xslx", or
        "/path/to/overnight.xslx".

    RETURNS

    list of commands
        *[command]* :
        List of command tuples for use in ``execute_command_list()``

    RAISES

    FileNotFoundError
        if file cannot be found

    SEE ALSO

    .. autosummary::

        ~get_command_list
        ~register_command_handler
        ~run_command_file
        ~summarize_command_file
        ~parse_text_command_file

    *new in apstools release 1.1.7*
    """
    full_filename = os.path.abspath(filename)
    if not os.path.exists(full_filename):
        raise FileNotFoundError(full_filename)
    xl = utils.ExcelDatabaseFileGeneric(full_filename)

    commands = []

    if len(xl.db) > 0:
        for i, row in enumerate(xl.db.values()):
            action, *values = list(row.values())

            # trim off any None values from end
            while len(values) > 0:
                if values[-1] is not None:
                    break
                values = values[:-1]

            commands.append((action, values, i + 1, list(row.values())))

    return commands


def parse_text_command_file(filename):
    """
    parse a text file with commands, return as command list

    * The text file is interpreted line-by-line.
    * Blank lines are ignored.
    * A pound sign (#) marks the rest of that line as a comment.
    * All remaining lines are interpreted as commands with arguments.

    Example of text file (no line numbers shown)::

        #List of sample scans to be run
        # pound sign starts a comment (through end of line)

        # action  value
        mono_shutter open

        # action  x y width height
        uslits 0 0 0.4 1.2

        # action  sx  sy  thickness   sample name
        FlyScan 0   0   0   blank
        FlyScan 5   2   0   "empty container"

        # action  sx  sy  thickness   sample name
        SAXS 0 0 0 blank

        # action  value
        mono_shutter close

    PARAMETERS

    filename
        *str* :
        Name of input text file.  Can be relative or absolute path,
        such as "actions.txt", "../sample.txt", or
        "/path/to/overnight.txt".

    RETURNS

    list of commands
        *[command]* :
        List of command tuples for use in ``execute_command_list()``

    RAISES

    FileNotFoundError
        if file cannot be found

    SEE ALSO

    .. autosummary::

        ~execute_command_list
        ~get_command_list
        ~register_command_handler
        ~run_command_file
        ~summarize_command_file
        ~parse_Excel_command_file

    *new in apstools release 1.1.7*
    """
    full_filename = os.path.abspath(filename)
    if not os.path.exists(full_filename):
        raise FileNotFoundError(full_filename)
    with open(full_filename, "r") as fp:
        buf = fp.readlines()

    commands = []
    for i, raw_command in enumerate(buf):
        row = raw_command.strip()
        if row == "" or row.startswith("#"):
            continue  # comment or blank
        else:  # command line
            action, *values = utils.split_quoted_line(row)
            commands.append((action, values, i + 1, raw_command.rstrip()))

    return commands


def register_command_handler(handler=None):
    """
    Define the function called to execute the command list

    PARAMETERS

    handler *obj* :
        Reference of the ``execute_command_list`` function
        to be used from :func:`~apstools.plans.run_command_file()`.
        If ``None`` or not provided,
        will reset to :func:`~apstools.plans.execute_command_list()`,
        which is also the initial setting.

    SEE ALSO

    .. autosummary::

        ~execute_command_list
        ~get_command_list
        ~register_command_handler
        ~summarize_command_file
        ~parse_Excel_command_file
        ~parse_text_command_file

    *new in apstools release 1.1.7*
    """
    COMMAND_LIST_REGISTRY.command = handler or execute_command_list


def run_command_file(filename, md=None):
    """
    plan: execute a list of commands from a text or Excel file

    .. index:: Bluesky Plan; run_command_file

    * Parse the file into a command list
    * yield the command list to the RunEngine (or other)

    SEE ALSO

    .. autosummary::

        ~execute_command_list
        ~get_command_list
        ~register_command_handler
        ~summarize_command_file
        ~parse_Excel_command_file
        ~parse_text_command_file

    *new in apstools release 1.1.7*
    """
    _md = dict(command_file=filename)
    _md.update(md or {})
    commands = get_command_list(filename)
    yield from COMMAND_LIST_REGISTRY.command(filename, commands, md=_md)


def summarize_command_file(filename):
    """
    print the command list from a text or Excel file

    SEE ALSO

    .. autosummary::

        ~execute_command_list
        ~get_command_list
        ~run_command_file
        ~parse_Excel_command_file
        ~parse_text_command_file

    *new in apstools release 1.1.7*
    """
    commands = get_command_list(filename)
    print(f"Command file: {filename}")
    print(command_list_as_table(commands))


COMMAND_LIST_REGISTRY = _HandlerRegistrar()
COMMAND_LIST_REGISTRY.command = execute_command_list

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Ophyd support for the EPICS synApps sscan record

see:  https://epics.anl.gov/bcda/synApps/sscan/SscanRecord.html

EXAMPLE::

    import apstools.synApps
    scans = apstools.synApps.SscanDevice("xxx:", name="scans")
    scans.select_channels()     # only the channels configured in EPICS


Public Structures

.. autosummary::

    ~SscanRecord
    ~SscanDevice

Private Structures

.. autosummary::

    ~sscanPositioner
    ~sscanDetector
    ~sscanTrigger

"""


from collections import OrderedDict
from ophyd import Device
from ophyd import Component as Cpt
from ophyd import DynamicDeviceComponent as DDC
from ophyd import FormattedComponent as FC
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO
from ophyd.status import DeviceStatus

from .. import utils as APS_utils


class sscanPositioner(Device):
    """
    positioner of an EPICS sscan record

    .. index:: Ophyd Device; synApps sscanPositioner

    .. autosummary::

        ~defined_in_EPICS
        ~reset

    """

    readback_pv = FC(EpicsSignal, "{self.prefix}.R{self._ch_num}PV", kind="config")
    readback_value = FC(EpicsSignalRO, "{self.prefix}.R{self._ch_num}CV", kind="hinted")
    array = FC(
        EpicsSignalRO,
        "{self.prefix}.P{self._ch_num}CA",
        kind="normal",  # TODO: which kind here?
    )
    setpoint_pv = FC(EpicsSignal, "{self.prefix}.P{self._ch_num}PV", kind="config")
    setpoint_value = FC(EpicsSignalRO, "{self.prefix}.P{self._ch_num}DV", kind="normal")
    start = FC(EpicsSignal, "{self.prefix}.P{self._ch_num}SP", kind="config")
    center = FC(EpicsSignal, "{self.prefix}.P{self._ch_num}CP", kind="config")
    end = FC(EpicsSignal, "{self.prefix}.P{self._ch_num}EP", kind="config")
    step_size = FC(EpicsSignal, "{self.prefix}.P{self._ch_num}SI", kind="config")
    width = FC(EpicsSignal, "{self.prefix}.P{self._ch_num}WD", kind="config")
    abs_rel = FC(EpicsSignal, "{self.prefix}.P{self._ch_num}AR", kind="config")
    mode = FC(EpicsSignal, "{self.prefix}.P{self._ch_num}SM", kind="config")
    units = FC(EpicsSignalRO, "{self.prefix}.P{self._ch_num}EU", kind="config")

    def __init__(self, prefix, num, **kwargs):
        self._ch_num = num
        super().__init__(prefix, **kwargs)

    def reset(self):
        """set all fields to default values"""
        self.readback_pv.put("")
        self.setpoint_pv.put("")
        self.start.put(0)
        self.center.put(0)
        self.end.put(0)
        self.step_size.put(0)
        self.width.put(0)
        self.abs_rel.put("ABSOLUTE")
        self.mode.put("LINEAR")

    @property
    def defined_in_EPICS(self):
        """True if defined in EPICS"""
        return len(self.setpoint_pv.get().strip()) > 0


class sscanDetector(Device):
    """
    detector of an EPICS sscan record

    .. index:: Ophyd Device; synApps sscanDetector

    .. autosummary::

        ~defined_in_EPICS
        ~reset

    """

    input_pv = FC(EpicsSignal, "{self.prefix}.D{self._ch_num}PV", kind="config")
    current_value = FC(EpicsSignal, "{self.prefix}.D{self._ch_num}CV", kind="hinted")
    array = FC(EpicsSignal, "{self.prefix}.D{self._ch_num}CA", kind="omitted")

    def __init__(self, prefix, num, **kwargs):
        self._ch_num = num
        super().__init__(prefix, **kwargs)

    def reset(self):
        """set all fields to default values"""
        self.input_pv.put("")

    @property
    def defined_in_EPICS(self):
        """True if defined in EPICS"""
        return len(self.input_pv.get().strip()) > 0


class sscanTrigger(Device):
    """
    detector trigger of an EPICS sscan record

    .. index:: Ophyd Device; synApps sscanTrigger

    .. autosummary::

        ~reset
    """

    trigger_pv = FC(EpicsSignal, "{self.prefix}.T{self._ch_num}PV", kind="config")
    trigger_value = FC(EpicsSignal, "{self.prefix}.T{self._ch_num}CD", kind="config")

    def __init__(self, prefix, num, **kwargs):
        self._ch_num = num
        super().__init__(prefix, **kwargs)

    def reset(self):
        """set all fields to default values"""
        self.trigger_pv.put("")
        self.trigger_value.put(1)

    @property
    def defined_in_EPICS(self):
        """True if defined in EPICS"""
        return len(self.trigger_pv.get().strip()) > 0


def _sscan_positioners(channel_list):
    defn = OrderedDict()
    for chan in channel_list:
        attr = "p{}".format(chan)
        defn[attr] = (sscanPositioner, "", {"num": chan})
    return defn


def _sscan_detectors(channel_list):
    defn = OrderedDict()
    for chan in channel_list:
        attr = "d{}".format(chan)
        defn[attr] = (sscanDetector, "", {"num": chan})
    return defn


def _sscan_triggers(channel_list):
    defn = OrderedDict()
    for chan in channel_list:
        attr = "t{}".format(chan)
        defn[attr] = (sscanTrigger, "", {"num": chan})
    return defn


class SscanRecord(Device):
    """
    EPICS synApps sscan record: used as ``$(P):scan(N)``

    .. index:: Ophyd Device; synApps SscanRecord

    .. autosummary::

        ~defined_in_EPICS
        ~reset
        ~select_channels

    """

    desc = Cpt(EpicsSignal, ".DESC", kind="config")
    scan_phase = Cpt(EpicsSignalRO, ".FAZE", kind="config")
    data_state = Cpt(EpicsSignalRO, ".DSTATE", kind="config")
    data_ready = Cpt(EpicsSignalRO, ".DATA", kind="config")
    scan_busy = Cpt(EpicsSignalRO, ".BUSY", kind="config")
    alert_flag = Cpt(EpicsSignalRO, ".ALRT", kind="config")
    alert_message = Cpt(EpicsSignalRO, ".SMSG", kind="config")
    number_points = Cpt(EpicsSignal, ".NPTS", kind="config")
    maximum_number_points = Cpt(EpicsSignal, ".MPTS", kind="config")
    current_point = Cpt(EpicsSignalRO, ".CPT", kind="normal")
    pasm = Cpt(EpicsSignal, ".PASM", kind="config")
    execute_scan = Cpt(EpicsSignal, ".EXSC", kind="omitted")
    bspv = Cpt(EpicsSignal, ".BSPV", kind="config")
    bscd = Cpt(EpicsSignal, ".BSCD", kind="omitted")
    bswait = Cpt(EpicsSignal, ".BSWAIT", kind="omitted")
    cmnd = Cpt(EpicsSignal, ".CMND", kind="omitted")
    detector_delay = Cpt(EpicsSignal, ".DDLY", kind="config")
    positioner_delay = Cpt(EpicsSignal, ".PDLY", kind="config")
    reference_detector = Cpt(EpicsSignal, ".REFD", kind="config")
    wait = Cpt(EpicsSignal, ".WAIT", kind="config")
    wcnt = Cpt(EpicsSignalRO, ".WCNT", kind="config")
    awct = Cpt(EpicsSignal, ".AWCT", kind="config")
    acqt = Cpt(EpicsSignal, ".ACQT", kind="config")
    acqm = Cpt(EpicsSignal, ".ACQM", kind="config")
    atime = Cpt(EpicsSignal, ".ATIME", kind="config")
    copyto = Cpt(EpicsSignal, ".COPYTO", kind="config")
    a1pv = Cpt(EpicsSignal, ".A1PV", kind="config")
    a1nv = Cpt(EpicsSignal, ".A1NV", kind="config")
    a1cd = Cpt(EpicsSignal, ".A1CD", kind="config")
    aspv = Cpt(EpicsSignal, ".ASPV", kind="config")
    ascd = Cpt(EpicsSignal, ".ASCD", kind="config")

    positioners = DDC(_sscan_positioners("1 2 3 4".split()))
    detectors = DDC(_sscan_detectors(APS_utils.itemizer("%02d", range(1, 71))))
    triggers = DDC(_sscan_triggers("1 2 3 4".split()))

    def set(self, value, **kwargs):
        """interface to use bps.mv()"""
        if value != 1:
            return

        working_status = DeviceStatus(self)
        started = False

        def execute_scan_cb(value, timestamp, **kwargs):
            value = int(value)
            if started and value == 0:
                working_status._finished()

        self.execute_scan.subscribe(execute_scan_cb)
        self.execute_scan.set(1)
        started = True
        return working_status

    def reset(self):
        """set all fields to default values"""
        self.desc.put(self.desc.pvname.split(".")[0])
        self.number_points.put(1000)
        for part in (self.positioners, self.detectors, self.triggers):
            for ch_name in part.component_names:
                channel = getattr(part, ch_name)
                channel.reset()
        self.a1pv.put("")
        self.acqm.put("NORMAL")
        if self.name.find("scanH") > 0:
            self.acqt.put("1D ARRAY")
        else:
            self.acqt.put("SCALAR")
        self.aspv.put("")
        self.bspv.put("")
        self.pasm.put("STAY")
        self.bswait.put("Wait")
        self.a1cd.put(1)
        self.ascd.put(1)
        self.bscd.put(1)
        self.reference_detector.put(1)
        self.atime.put(0)
        self.awct.put(0)
        self.copyto.put(0)
        self.detector_delay.put(0)
        self.positioner_delay.put(0)
        while self.wcnt.get() > 0:
            self.wait.put(0)

    def select_channels(self):
        """
        Select channels that are configured in EPICS
        """
        for part in (self.positioners, self.detectors, self.triggers):
            channel_names = []  # part.get_configured_channels()
            # fmt: off
            channel_names = [
                ch
                for ch in part.component_names
                if getattr(part, ch).defined_in_EPICS
            ]
            # fmt: on

            part.configuration_attrs = channel_names
            part.read_attrs = channel_names
            part.kind = "normal"

    @property
    def defined_in_EPICS(self):
        """True if will be used in EPICS"""
        self.select_channels()
        channels = len(self.positioners.read_attrs)
        channels += len(self.detectors.read_attrs)
        # channels += len(self.triggers.read_attrs)
        return channels > 0


class SscanDevice(Device):
    """
    EPICS synApps XXX IOC setup of sscan records: ``$(P):scan$(N)``

    .. index:: Ophyd Device; synApps SscanDevice

    .. autosummary::

        ~reset
        ~select_channels

    """

    scan_dimension = Cpt(EpicsSignalRO, "ScanDim", kind="config")
    scan_pause = Cpt(EpicsSignal, "scanPause", kind="omitted")
    abort_scans = Cpt(EpicsSignal, "AbortScans", kind="omitted")
    scan1 = Cpt(SscanRecord, "scan1")
    scan2 = Cpt(SscanRecord, "scan2")
    scan3 = Cpt(SscanRecord, "scan3")
    scan4 = Cpt(SscanRecord, "scan4")
    scanH = Cpt(SscanRecord, "scanH")
    resume_delay = Cpt(EpicsSignal, "scanResumeSEQ.DLY1", kind="config")

    def reset(self):
        """set all fields to default values"""
        for chnum in "1 2 3 4 H".split():
            getattr(self, "scan" + chnum).reset()

    def select_channels(self):
        """
        Select only the scans that are configured in EPICS
        """
        scans = ["scan" + ch for ch in "1 2 3 4 H".split()]
        attrs = []
        for nm in self.component_names:
            if nm in scans and not getattr(self, nm).defined_in_EPICS:
                continue
            attrs.append(nm)
        self.read_attrs = attrs

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
email Support
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~EmailNotifications
"""

import smtplib
from email.mime.text import MIMEText


class EmailNotifications(object):
    """
    send email notifications when requested

    .. index:: apstools Utility; EmailNotifications

    use default OS mail utility (so no credentials needed)

    EXAMPLE

    Send email(s) when `feedback_limits_approached`
    (a hypothetical boolean) is `True`::

        # setup
        from apstools.utils import EmailNotifications

        SENDER_EMAIL = "instrument_user@email.host.tld"

        email_notices = EmailNotifications(SENDER_EMAIL)
        email_notices.add_addresses(
            # This list receives email when send() is called.
            "joe.user@goodmail.com",
            "instrument_team@email.host.tld",
            # others?
        )

        # ... later

        if feedback_limits_approached:
            # send emails to list
            subject = "Feedback problem"
            message = "Feedback is very close to its limits."
            email_notices.send(subject, message)
    """
    from .misc import run_in_thread

    def __init__(self, sender=None):
        self.addresses = []
        self.notify_on_feedback = True
        self.sender = sender or "nobody@localhost"
        self.smtp_host = "localhost"

    def add_addresses(self, *args):
        for address in args:
            self.addresses.append(address)

    @run_in_thread
    def send(self, subject, message):
        """send ``message`` to all addresses"""
        msg = MIMEText(message)
        msg["Subject"] = subject
        msg["From"] = self.sender
        msg["To"] = ",".join(self.addresses)
        s = smtplib.SMTP(self.smtp_host)
        s.sendmail(self.sender, self.addresses, msg.as_string())
        s.quit()

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
nscan plan
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~nscan
"""

from collections import OrderedDict
import datetime
import numpy as np

from bluesky import plan_stubs as bps
from bluesky import preprocessors as bpp


def nscan(detectors, *motor_sets, num=11, per_step=None, md=None):
    """
    Scan over ``n`` variables moved together, each in equally spaced steps.

    .. index:: Bluesky Plan; nscan

    PARAMETERS

    detectors *list* :
        list of 'readable' objects
    motor_sets *list* :
        sequence of one or more groups of: motor, start, finish
    motor *object* :
        any 'settable' object (motor, temp controller, etc.)
    start *float* :
        starting position of motor
    finish *float* :
        ending position of motor
    num *int* :
        number of steps (default = 11)
    per_step *callable* :
        (optional)
        hook for customizing action of inner loop (messages per step)
        Expected signature: ``f(detectors, step_cache, pos_cache)``
    md *dict*
        (optional)
        metadata

    See the ``nscan()`` example in a Jupyter notebook:
    https://github.com/BCDA-APS/apstools/blob/master/docs/source/resources/demo_nscan.ipynb
    """

    def take_n_at_a_time(args, n=2):
        yield from zip(*[iter(args)] * n)

    if len(motor_sets) < 3:
        raise ValueError("must provide at least one movable")
    if len(motor_sets) % 3 > 0:
        raise ValueError("must provide sets of movable, start, finish")

    motors = OrderedDict()
    for m, s, f in take_n_at_a_time(motor_sets, n=3):
        if not isinstance(s, (int, float)):
            msg = "start={} ({}): is not a number".format(s, type(s))
            raise ValueError(msg)
        if not isinstance(f, (int, float)):
            msg = "finish={} ({}): is not a number".format(f, type(f))
            raise ValueError(msg)
        motors[m.name] = dict(
            motor=m,
            start=s,
            finish=f,
            steps=np.linspace(start=s, stop=f, num=num),
        )

    _md = {
        "detectors": [det.name for det in detectors],
        "motors": [m for m in motors.keys()],
        "num_points": num,
        "num_intervals": num - 1,
        "plan_args": {
            "detectors": list(map(repr, detectors)),
            "num": num,
            "motors": repr(motor_sets),
            "per_step": repr(per_step),
        },
        "plan_name": "nscan",
        "plan_pattern": "linspace",
        "hints": {},
        "iso8601": datetime.datetime.now(),
    }
    _md.update(md or {})

    try:
        m = list(motors.keys())[0]
        dimensions = [(motors[m]["motor"].hints["fields"], "primary")]
    except (AttributeError, KeyError):
        pass
    else:
        _md["hints"].setdefault("dimensions", dimensions)

    if per_step is None:
        per_step = bps.one_nd_step

    @bpp.stage_decorator(
        list(detectors) + [m["motor"] for m in motors.values()]
    )
    @bpp.run_decorator(md=_md)
    def inner_scan():
        for step in range(num):
            step_cache, pos_cache = {}, {}
            for m in motors.values():
                next_pos = m["steps"][step]
                m = m["motor"]
                pos_cache[m] = m.read()[m.name]["value"]
                step_cache[m] = next_pos
            yield from per_step(detectors, step_cache, pos_cache)

    return (yield from inner_scan())

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Shutters
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~ApsPssShutter
   ~ApsPssShutterWithStatus
   ~EpicsMotorShutter
   ~EpicsOnOffShutter
   ~OneSignalShutter
   ~ShutterBase
   ~SimulatedApsPssShutterWithStatus
"""

from ophyd import Component
from ophyd import Device
from ophyd import DeviceStatus
from ophyd import EpicsMotor
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO
from ophyd import FormattedComponent
from ophyd import Signal

import numpy as np
import threading
import time


class ShutterBase(Device):
    """
    Base class for all shutter Devices.

    .. index:: Ophyd Device; ShutterBase

    PARAMETERS

    value
        *str* :
        any from ``self.choices`` (typically "open" or "close")

    valid_open_values
        *[str]* :
        A list of lower-case text values that are acceptable
        for use with the ``set()`` command to open the shutter.

    valid_close_values
        *[str]* :
        A list of lower-case text values that are acceptable
        for use with the ``set()`` command to close the shutter.

    open_value
        *number* :
        The actual value to send to open ``signal`` to open the shutter.
        (default = 1)

    close_value
        *number* :
        The actual value to send to close ``signal`` to close the shutter.
        (default = 0)

    delay_s
        *float* :
        time to wait (s) after move is complete,
        does not wait if shutter already in position
        (default = 0)

    busy
        *Signal* :
        (internal) tells if a move is in progress

    unknown_state
        *str* :
        (constant) Text reported by ``state`` when not open or closed.
        cannot move to this position
        (default = "unknown")
    """

    # fmt: off
    valid_open_values = ["open", "opened"]  # lower-case strings ONLY
    valid_close_values = ["close", "closed"]
    # fmt: on
    open_value = 1  # value of "open"
    close_value = 0  # value of "close"
    delay_s = 0.0  # time to wait (s) after move is complete
    busy = Component(Signal, value=False)
    unknown_state = "unknown"  # cannot move to this position

    # - - - - likely to override these methods in subclass - - - -

    def open(self):
        """
        BLOCKING: request shutter to open, called by ``set()``.

        Must implement in subclass of ShutterBase()

        EXAMPLE::

            if not self.isOpen:
                self.signal.put(self.open_value)
                if self.delay_s > 0:
                    time.sleep(self.delay_s)    # blocking call OK here

        """
        raise NotImplementedError("must implement in subclass")

    def close(self):
        """
        BLOCKING: request shutter to close, called by ``set()``.

        Must implement in subclass of ShutterBase()

        EXAMPLE::

            if not self.isClosed:
                self.signal.put(self.close_value)
                if self.delay_s > 0:
                    time.sleep(self.delay_s)    # blocking call OK here

        """
        raise NotImplementedError("must implement in subclass")

    @property
    def state(self):
        """
        returns ``open``, ``close``, or ``unknown``

        Must implement in subclass of ShutterBase()

        EXAMPLE::

            if self.signal.get() == self.open_value:
                result = self.valid_open_values[0]
            elif self.signal.get() == self.close_value:
                result = self.valid_close_values[0]
            else:
                result = self.unknown_state
            return result

        """
        raise NotImplementedError("must implement in subclass")

    # - - - - - - possible to override in subclass - - - - - -

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.valid_open_values = list(map(self.lowerCaseString, self.valid_open_values))
        self.valid_close_values = list(
            map(self.lowerCaseString, self.valid_close_values)
        )

    @property
    def isOpen(self):
        """is the shutter open?"""
        return str(self.state) == self.valid_open_values[0]

    @property
    def isClosed(self):
        """is the shutter closed?"""
        return str(self.state) == self.valid_close_values[0]

    def inPosition(self, target):
        """is the shutter at the target position?"""
        self.validTarget(target)
        __value__ = self.lowerCaseString(target)
        if __value__ in self.valid_open_values and self.isOpen:
            return True
        elif __value__ in self.valid_close_values and self.isClosed:
            return True
        return False

    def set(self, value, **kwargs):
        """
        plan: request the shutter to open or close

        PARAMETERS

        value
            *str* :
            any from ``self.choices`` (typically "open" or "close")

        kwargs
            *dict* :
            ignored at this time

        """
        if self.busy.get():
            raise RuntimeError("shutter is operating")

        __value__ = self.lowerCaseString(value)
        self.validTarget(__value__)

        status = DeviceStatus(self)

        if self.inPosition(__value__):
            # no need to move, cut straight to the end
            status._finished(success=True)
        else:

            def move_it():
                # runs in a thread, no need to "yield from"
                self.busy.put(True)
                if __value__ in self.valid_open_values:
                    self.open()
                elif __value__ in self.valid_close_values:
                    self.close()
                self.busy.put(False)
                status._finished(success=True)

            # get it moving
            threading.Thread(target=move_it, daemon=True).start()
        return status

    # - - - - - - not likely to override in subclass - - - - - -

    def addCloseValue(self, text):
        """a synonym to close the shutter, use with set()"""
        self.valid_close_values.append(self.lowerCaseString(text))
        return self.choices  # return the list of acceptable values

    def addOpenValue(self, text):
        """a synonym to open the shutter, use with set()"""
        self.valid_open_values.append(self.lowerCaseString(text))
        return self.choices  # return the list of acceptable values

    @property
    def choices(self):
        """return list of acceptable choices for set()"""
        return self.valid_open_values + self.valid_close_values

    def lowerCaseString(self, value):
        """ensure any given value is a lower-case string"""
        return str(value).lower()

    def validTarget(self, target, should_raise=True):
        """
        return whether (or not) target value is acceptable for self.set()

        raise ValueError if not acceptable (default)
        """
        acceptable_values = self.choices
        ok = self.lowerCaseString(target) in acceptable_values
        if not ok and should_raise:
            msg = "received " + str(target)
            msg += " : should be only one of "
            msg += " | ".join(acceptable_values)
            raise ValueError(msg)
        return ok


class OneSignalShutter(ShutterBase):
    """
    Shutter Device using one Signal for open and close.

    .. index:: Ophyd Device; OneSignalShutter

    PARAMETERS

    signal
        ``EpicsSignal`` or ``Signal`` :
        (override in subclass)
        The ``signal`` is the comunication to the hardware.
        In a subclass, the hardware may have more than
        one communication channel to use.  See the
        ``ApsPssShutter`` as an example.

    See ``ShutterBase`` for more parameters.

    EXAMPLE

    Create a simulated shutter:

        shutter = OneSignalShutter(name="shutter")

    open the shutter (interactively):

        shutter.open()

    Check the shutter is open:

        In [144]: shutter.isOpen
        Out[144]: True

    Use the shutter in a Bluesky plan.
    Set a post-move delay time of 1.0 seconds.
    Be sure to use ``yield from``, such as::

        def in_a_plan(shutter):
            shutter.delay_s = 1.0
            t0 = time.time()
            print("Shutter state: " + shutter.state, time.time()-t0)
            yield from bps.abs_set(shutter, "open", wait=True)    # wait for completion is optional
            print("Shutter state: " + shutter.state, time.time()-t0)
            yield from bps.mv(shutter, "open")    # do it again
            print("Shutter state: " + shutter.state, time.time()-t0)
            yield from bps.mv(shutter, "close")    # ALWAYS waits for completion
            print("Shutter state: " + shutter.state, time.time()-t0)

        RE(in_a_plan(shutter))

    which gives this output:

        Shutter state: close 1.7642974853515625e-05
        Shutter state: open 1.0032124519348145
        Shutter state: open 1.0057861804962158
        Shutter state: close 2.009695529937744

    The strings accepted by `set()` are defined in two lists:
    `valid_open_values` and `valid_close_values`.  These lists
    are treated (internally to `set()`) as lower case strings.

    Example, add "o" & "x" as aliases for "open" & "close":

        shutter.addOpenValue("o")
        shutter.addCloseValue("x")
        shutter.set("o")
        shutter.set("x")
    """

    signal = Component(Signal, value=0)

    @property
    def state(self):
        """is shutter "open", "close", or "unknown"?"""
        if self.signal.get() == self.open_value:
            result = self.valid_open_values[0]
        elif self.signal.get() == self.close_value:
            result = self.valid_close_values[0]
        else:
            result = self.unknown_state
        return result

    def open(self):
        """BLOCKING: request shutter to open, called by set()"""
        if not self.isOpen:
            self.signal.put(self.open_value)
            if self.delay_s > 0:
                time.sleep(self.delay_s)  # blocking call OK here

    def close(self):
        """BLOCKING: request shutter to close, called by set()"""
        if not self.isClosed:
            self.signal.put(self.close_value)
            if self.delay_s > 0:
                time.sleep(self.delay_s)  # blocking call OK here


class ApsPssShutter(ShutterBase):
    """
    APS PSS shutter

    .. index:: Ophyd Device; ApsPssShutter

    * APS PSS shutters have separate bit PVs for open and close
    * set either bit, the shutter moves, and the bit resets a short time later
    * no indication that the shutter has actually moved from the bits
      (see :func:`ApsPssShutterWithStatus()` for alternative)

    Since there is no direct indication that a shutter has moved, the
    ``state`` property will always return *unknown* and the
    ``isOpen`` and ``isClosed`` properties will always return *False*.

    A consequence of the unknown state is that the shutter will always
    be commanded to move (and wait the ``delay_s`` time),
    even if it is already at that position.  This device could keep
    track of the last commanded position, but that is not guaranteed
    to be true since the shutter could be moved from other software.

    The default ``delay_s`` has been set at *1.2 s* to allow for
    shutter motion.  Change this as desired.  Advise if this
    default should be changed.

    EXAMPLE::

        shutter_a = ApsPssShutter("2bma:A_shutter:", name="shutter")

        shutter_a.open()
        shutter_a.close()

        shutter_a.set("open")
        shutter_a.set("close")

    When using the shutter in a plan, be sure to use ``yield from``, such as::

        def in_a_plan(shutter):
            yield from abs_set(shutter, "open", wait=True)
            # do something
            yield from abs_set(shutter, "close", wait=True)

        RE(in_a_plan(shutter_a))

    The strings accepted by `set()` are defined in two lists:
    `valid_open_values` and `valid_close_values`.  These lists
    are treated (internally to `set()`) as lower case strings.

    Example, add "o" & "x" as aliases for "open" & "close":

        shutter_a.addOpenValue("o")
        shutter_a.addCloseValue("x")
        shutter_a.set("o")
        shutter_a.set("x")
    """

    # bo records that reset after a short time, set to 1 to move
    # note: upper-case first characters here (unique to 9-ID)?
    open_signal = Component(EpicsSignal, "Open")
    close_signal = Component(EpicsSignal, "Close")

    delay_s = 1.2  # allow time for shutter to move

    @property
    def state(self):
        """is shutter "open", "close", or "unknown"?"""
        return self.unknown_state  # no state info available

    def open(self, timeout=10):
        """request the shutter to open (timeout is ignored)"""
        if not self.isOpen:
            self.open_signal.put(1)

            # wait for the shutter to move
            if self.delay_s > 0:
                time.sleep(self.delay_s)  # blocking call OK here

            # reset that signal (if not done by EPICS)
            if self.open_signal.get() == 1:
                self.open_signal.put(0)

    def close(self, timeout=10):
        """request the shutter to close (timeout is ignored)"""
        if not self.isClosed:
            self.close_signal.put(1)

            # wait for the shutter to move
            if self.delay_s > 0:
                time.sleep(self.delay_s)  # blocking call OK here

            # reset that signal (if not done by EPICS)
            if self.close_signal.get() == 1:
                self.close_signal.put(0)


class ApsPssShutterWithStatus(ApsPssShutter):
    """
    APS PSS shutter with separate status PV

    .. index:: Ophyd Device; ApsPssShutterWithStatus

    * APS PSS shutters have separate bit PVs for open and close
    * set either bit, the shutter moves, and the bit resets a short time later
    * a separate status PV tells if the shutter is open or closed
      (see :func:`ApsPssShutter()` for alternative)

    EXAMPLE::

        A_shutter = ApsPssShutterWithStatus(
            "2bma:A_shutter:",
            "PA:02BM:STA_A_FES_OPEN_PL",
            name="A_shutter")
        B_shutter = ApsPssShutterWithStatus(
            "2bma:B_shutter:",
            "PA:02BM:STA_B_SBS_OPEN_PL",
            name="B_shutter")

        A_shutter.open()
        A_shutter.close()

        or

        A_shutter.set("open")
        A_shutter.set("close")

    When using the shutter in a plan, be sure to use `yield from`.

        def in_a_plan(shutter):
            yield from abs_set(shutter, "open", wait=True)
            # do something
            yield from abs_set(shutter, "close", wait=True)

        RE(in_a_plan(A_shutter))

    """

    # bi record ZNAM=OFF, ONAM=ON
    pss_state = FormattedComponent(EpicsSignalRO, "{self.state_pv}")
    pss_state_open_values = [1]
    pss_state_closed_values = [0]

    delay_s = 0  # let caller add time after the move

    _poll_factor_ = 1.5
    _poll_s_min_ = 0.002
    _poll_s_max_ = 0.15

    def __init__(self, prefix, state_pv, *args, **kwargs):
        self.state_pv = state_pv
        super().__init__(prefix, *args, **kwargs)

    @property
    def state(self):
        """is shutter "open", "close", or "unknown"?"""
        # update the list of acceptable values - very inefficient but works
        for item in self.pss_state.enum_strs[1]:
            if item not in self.pss_state_open_values:
                self.pss_state_open_values.append(item)
        for item in self.pss_state.enum_strs[0]:
            if item not in self.pss_state_closed_values:
                self.pss_state_closed_values.append(item)

        if self.pss_state.get() in self.pss_state_open_values:
            result = self.valid_open_values[0]
        elif self.pss_state.get() in self.pss_state_closed_values:
            result = self.valid_close_values[0]
        else:
            result = self.unknown_state
        return result

    def wait_for_state(self, target, timeout=10, poll_s=0.01):
        """
        wait for the PSS state to reach a desired target

        PARAMETERS

        target
            *[str]* :
            list of strings containing acceptable values

        timeout
            *non-negative number* :
            maximum amount of time (seconds) to wait for PSS state to reach target

        poll_s
            *non-negative number* :
            Time to wait (seconds) in first polling cycle.
            After first poll, this will be increased by ``_poll_factor_``
            up to a maximum time of ``_poll_s_max_``.
        """
        if timeout is not None:
            expiration = time.time() + max(timeout, 0)  # ensure non-negative timeout
        else:
            expiration = None

        # ensure the poll delay is reasonable
        if poll_s > self._poll_s_max_:
            poll_s = self._poll_s_max_
        elif poll_s < self._poll_s_min_:
            poll_s = self._poll_s_min_

        while self.pss_state.get() not in target:
            time.sleep(poll_s)
            if poll_s < self._poll_s_max_:
                poll_s *= self._poll_factor_  # progressively longer
            if expiration is not None and time.time() > expiration:
                msg = f"Timeout ({timeout} s) waiting for shutter state"
                msg += f" to reach a value in {target}"
                raise TimeoutError(msg)

    def open(self, timeout=10):
        """request the shutter to open"""
        if not self.isOpen:
            self.open_signal.put(1)

            # wait for the shutter to move
            self.wait_for_state(self.pss_state_open_values, timeout=timeout)

            # wait as caller specified
            if self.delay_s > 0:
                time.sleep(self.delay_s)  # blocking call OK here

            # reset that signal (if not done by EPICS)
            if self.open_signal.get() == 1:
                self.open_signal.put(0)

    def close(self, timeout=10):
        """request the shutter to close"""
        if not self.isClosed:
            self.close_signal.put(1)

            # wait for the shutter to move
            self.wait_for_state(self.pss_state_closed_values, timeout=timeout)

            # wait as caller specified
            if self.delay_s > 0:
                time.sleep(self.delay_s)  # blocking call OK here

            # reset that signal (if not done by EPICS)
            if self.close_signal.get() == 1:
                self.close_signal.put(0)


class SimulatedApsPssShutterWithStatus(ApsPssShutterWithStatus):
    """
    Simulated APS PSS shutter

    .. index:: Ophyd Device; SimulatedApsPssShutterWithStatus

    EXAMPLE::

        sim = SimulatedApsPssShutterWithStatus(name="sim")

    """

    open_signal = Component(Signal, value=0)
    close_signal = Component(Signal, value=0)
    pss_state = FormattedComponent(Signal, value="close")

    def __init__(self, *args, **kwargs):
        # was: super(ApsPssShutter, self).__init__("", *args, **kwargs)
        super(SimulatedApsPssShutterWithStatus, self).__init__("", "", *args, **kwargs)
        self.pss_state_open_values += self.valid_open_values
        self.pss_state_closed_values += self.valid_close_values

    def wait_for_state(self, target, timeout=10, poll_s=0.01):
        """
        wait for the PSS state to reach a desired target

        PARAMETERS

        target
            *[str]* :
            list of strings containing acceptable values

        timeout
            *non-negative number* :
            Ignored in the simulation.

        poll_s
            *non-negative number* :
            Ignored in the simulation.
        """
        simulated_response_time_s = np.random.uniform(0.1, 0.9)
        time.sleep(simulated_response_time_s)
        self.pss_state.put(target[0])

    @property
    def state(self):
        """is shutter "open", "close", or "unknown"?"""
        if self.pss_state.get() in self.pss_state_open_values:
            result = self.valid_open_values[0]
        elif self.pss_state.get() in self.pss_state_closed_values:
            result = self.valid_close_values[0]
        else:
            result = self.unknown_state
        return result


class EpicsMotorShutter(OneSignalShutter):
    """
    Shutter, implemented with an EPICS motor moved between two positions

    .. index:: Ophyd Device; EpicsMotorShutter

    EXAMPLE::

        tomo_shutter = EpicsMotorShutter("2bma:m23", name="tomo_shutter")
        tomo_shutter.close_value = 1.0      # default
        tomo_shutter.open_value = 0.0       # default
        tomo_shutter.tolerance = 0.01       # default
        tomo_shutter.open()
        tomo_shutter.close()

        # or, when used in a plan
        def planA():
            yield from abs_set(tomo_shutter, "open", group="O")
            yield from wait("O")
            yield from abs_set(tomo_shutter, "close", group="X")
            yield from wait("X")
        def planA():
            yield from abs_set(tomo_shutter, "open", wait=True)
            yield from abs_set(tomo_shutter, "close", wait=True)
        def planA():
            yield from mv(tomo_shutter, "open")
            yield from mv(tomo_shutter, "close")

    """

    signal = Component(EpicsMotor, "")
    tolerance = 0.01  # how close is considered in-position?

    @property
    def state(self):
        """is shutter "open", "close", or "unknown"?"""
        if abs(self.signal.user_readback.get() - self.open_value) <= self.tolerance:
            result = self.valid_open_values[0]
        elif abs(self.signal.user_readback.get() - self.close_value) <= self.tolerance:
            result = self.valid_close_values[0]
        else:
            result = self.unknown_state
        return result

    def open(self):
        """move motor to BEAM NOT BLOCKED position, interactive use"""
        if not self.isOpen:
            self.signal.move(self.open_value)
            if self.delay_s > 0:
                time.sleep(self.delay_s)  # blocking call OK here

    def close(self):
        """move motor to BEAM BLOCKED position, interactive use"""
        self.signal.move(self.close_value)
        if not self.isClosed:
            self.signal.move(self.close_value)
            if self.delay_s > 0:
                time.sleep(self.delay_s)  # blocking call OK here


class EpicsOnOffShutter(OneSignalShutter):
    """
    Shutter using a single EPICS PV moved between two positions

    .. index:: Ophyd Device; EpicsOnOffShutter

    Use for a shutter controlled by a single PV which takes a
    value for the close command and a different value for the open command.
    The current position is determined by comparing the value of the control
    with the expected open and close values.

    EXAMPLE::

        bit_shutter = EpicsOnOffShutter("2bma:bit1", name="bit_shutter")
        bit_shutter.close_value = 0      # default
        bit_shutter.open_value = 1       # default
        bit_shutter.open()
        bit_shutter.close()

        # or, when used in a plan
        def planA():
            yield from mv(bit_shutter, "open")
            yield from mv(bit_shutter, "close")

    """

    signal = Component(EpicsSignal, "")

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
APS cycles
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~ApsCycleDM
"""

import datetime
import json
from ophyd.sim import SynSignalRO
import pathlib
import time
import yaml


_PATH = pathlib.Path(__file__).parent
YAML_CYCLE_FILE = (_PATH / "aps_cycle_info.yml")


class _ApsCycleDB:
    """
    Python representation of the APS run cycle schedule table.
    """

    def __init__(self):
        self.db = self._read_cycle_data()

    def get_cycle_name(self, ts=None):
        """
        Get the name of the current APS run cycle.

        By default, the name of the current run cycle (based on the
        current timestamp) will be returned.

        PARAMETERS

        ts float:
            Absolute time stamp (such as from ``time.time()``).
            Default: current time stamp.

        RETURNS

        Returns cycle name (str) or ``None`` if timestamp is not in data table.
        """
        ts = ts or time.time()
        for cycle, span in self.db.items():
            if span["start"] <= ts < span["end"]:
                return cycle
        return None  # not found

    def _read_cycle_data(self):
        """
        Read the list of APS run cycles from a local file.

        The file is formatted in YAML after reformatting content received from
        the APS Data Management package (*aps-dm-api*).  The YAML format is
        easily updated and human-readable.
        """
        _cycles_yml = YAML_CYCLE_FILE.open().read()
        _cycles = yaml.load(_cycles_yml, Loader=yaml.BaseLoader)

        def iso2ts(isodatetime):
            return datetime.datetime.timestamp(
                datetime.datetime.fromisoformat(isodatetime)
            )

        db = {
            run_name: dict(
                start=iso2ts(span["begin"]), end=iso2ts(span["end"])
            )
            for run_name, span in _cycles.items()
        }
        return db

    def _write_cycle_data(self):
        """
        Write the list of APS run cycles to a local file.

        The file is formatted exactly as received from the
        APS Data Management package (*aps-dm-api*).  This allows
        automatic updates as needed.

        To update the LOCAL_FILE, run this code (on a workstation at the APS)::

            from apstools._devices.aps_cycle import cycle_db
            cycle_db._write_cycle_data()
        """
        runs = self._bss_list_runs
        if runs is not None:
            cycle_data = runs.replace("'", '"')
            cycles = json.loads(cycle_data)

            def sorter(line):
                return line["name"]

            db = {
                line["name"]: {
                    "begin": line["startTime"],
                    "end": line["endTime"],
                }
                for line in sorted(cycles, key=sorter)
            }
            with YAML_CYCLE_FILE.open("w") as f:
                f.write(yaml.dump(db))

    @property
    def _bss_list_runs(self):
        try:
            import apsbss

            return str(apsbss.api_bss.listRuns())
        except ModuleNotFoundError:
            return None


cycle_db = _ApsCycleDB()


class ApsCycleDM(SynSignalRO):
    """
    Get the APS cycle name from the APS Data Management system or a local file.

    .. index:: Ophyd Signal; ApsCycleDM

    This signal is read-only.
    """

    _cycle_ends = "1980"  # force a read from DM on first get()
    _cycle_name = "unknown"

    def get(self):
        self._cycle_name = cycle_db.get_cycle_name()
        if datetime.datetime.now().isoformat(sep=" ") >= self._cycle_ends:
            self._cycle_ends = datetime.datetime.fromtimestamp(
                cycle_db.db[self._cycle_name]["end"]
            ).isoformat(sep=" ")
        return self._cycle_name

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



from .. import EpicsScanIdSignal
from ...tests import IOC
from ophyd import EpicsSignal
import bluesky
import bluesky.plans as bp
import pytest

SCAN_ID_PV = f"{IOC}gp:int1"


@pytest.mark.parametrize(
    "pvname, initial_value",
    [
        [None, 0],
        [SCAN_ID_PV, 0],
        [SCAN_ID_PV, 101],
    ],
)
def test_usage(pvname, initial_value):
    if pvname is None:
        scan_id_epics = None
        callback = None
    else:
        scan_id_epics = EpicsScanIdSignal(pvname, name="scan_id_epics")
        scan_id_epics.wait_for_connection()
        scan_id_epics.put(initial_value)
        callback = scan_id_epics.cb_scan_id_source

    if scan_id_epics is None:
        RE = bluesky.RunEngine({})
    else:
        RE = bluesky.RunEngine({}, scan_id_source=callback)
    assert isinstance(RE.md, dict)
    assert "scan_id" not in RE.md

    signal = EpicsSignal(f"{IOC}gp:float1", name="signal")
    signal.wait_for_connection()

    RE(bp.count([signal]))
    assert "scan_id" in RE.md
    assert RE.md["scan_id"] == initial_value + 1



"""
Ophyd support for EPICS synApps modules (records and databases).
"""

from ._common import EpicsRecordDeviceCommonAll
from ._common import EpicsRecordInputFields
from ._common import EpicsRecordOutputFields
from ._common import EpicsRecordFloatFields
from ._common import EpicsSynAppsRecordEnableMixin

from .asyn import AsynRecord

from .busy import BusyRecord

from .calcout import CalcoutRecord
from .calcout import CalcoutRecordChannel
from .calcout import setup_gaussian_calcout
from .calcout import setup_incrementer_calcout
from .calcout import setup_lorentzian_calcout
from .calcout import UserCalcoutN
from .calcout import UserCalcoutDevice

from .epid import EpidRecord

from .iocstats import IocStatsDevice

from .luascript import LuascriptRecord
from .luascript import LuascriptRecordNumberInput
from .luascript import LuascriptRecordStringInput
from .luascript import UserScriptsDevice

from .save_data import SaveData

from .scalcout import UserScalcoutDevice
from .scalcout import UserScalcoutN
from .scalcout import ScalcoutRecord
from .scalcout import ScalcoutRecordNumberChannel
from .scalcout import ScalcoutRecordStringChannel

from .sscan import SscanRecord
from .sscan import SscanDevice

from .sseq import EditStringSequence
from .sseq import SseqRecord
from .sseq import UserStringSequenceDevice
from .sseq import UserStringSequenceN

from .sub import SubRecord
from .sub import SubRecordChannel
from .sub import UserAverageN
from .sub import UserAverageDevice

from .swait import SwaitRecord
from .swait import SwaitRecordChannel
from .swait import UserCalcN
from .swait import UserCalcsDevice
from .swait import setup_random_number_swait
from .swait import setup_gaussian_swait
from .swait import setup_lorentzian_swait
from .swait import setup_incrementer_swait

from .transform import TransformRecord
from .transform import UserTransformN
from .transform import UserTransformsDevice

# MUST come AFTER previous imports
from .db_2slit import Optics2Slit1D
from .db_2slit import Optics2Slit2D_HV
from .db_2slit import Optics2Slit2D_InbOutBotTop

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



import pytest

from ..swait import setup_random_number_swait
from ..swait import SwaitRecord
from ..swait import UserCalcN
from ..swait import UserCalcsDevice
from ...tests import common_attribute_quantities_test
from ...tests import IOC
from ...tests import short_delay_for_EPICS_IOC_database_processing


@pytest.mark.parametrize(
    "device, pv, connect, attr, expected",
    [
        [SwaitRecord, f"{IOC}userCalc10", False, "read_attrs", 12],
        [SwaitRecord, f"{IOC}userCalc10", False, "configuration_attrs", 72],
        [SwaitRecord, f"{IOC}userCalc10", True, "read()", 1],
        [SwaitRecord, f"{IOC}userCalc10", True, "summary()", 162],

        [UserCalcN, f"{IOC}userCalc10", False, "read_attrs", 12],
        [UserCalcN, f"{IOC}userCalc10", False, "configuration_attrs", 73],
        [UserCalcN, f"{IOC}userCalc10", True, "read()", 1],
        [UserCalcN, f"{IOC}userCalc10", True, "summary()", 164],

        [UserCalcsDevice, IOC, False, "read_attrs", 130],
        [UserCalcsDevice, IOC, False, "configuration_attrs", 741],
        [UserCalcsDevice, IOC, True, "read()", 10],
        [UserCalcsDevice, IOC, True, "summary()", 1496],
    ]
)
def test_attribute_quantities(device, pv, connect, attr, expected):
    """Verify the quantities of the different attributes."""
    common_attribute_quantities_test(device, pv, connect, attr, expected)


def test_userCalcs_reset():
    user = UserCalcsDevice(IOC, name="user")
    user.wait_for_connection()
    user.enable.put("Enable")

    swait = user.calc10
    assert isinstance(swait, SwaitRecord)
    swait.enable.put("E")  # Note: only "E"

    setup_random_number_swait(swait)
    short_delay_for_EPICS_IOC_database_processing(0.2)
    assert swait.description.get() == "uniform random numbers"
    assert swait.calculation.get() == "RNDM"
    v1 = swait.calculated_value.get()
    short_delay_for_EPICS_IOC_database_processing(0.2)
    assert v1 != swait.calculated_value.get()

    swait.reset()
    short_delay_for_EPICS_IOC_database_processing()
    assert swait.description.get() == swait.prefix
    assert swait.calculation.get() == "0"
    v1 = swait.calculated_value.get()
    short_delay_for_EPICS_IOC_database_processing(0.2)
    assert v1 == swait.calculated_value.get()



"""
Request user input in a plan
++++++++++++++++++++++++++++

.. autosummary::

   ~request_input

Wrap ``bps.input_plan()`` to *ask* the user a question.
"""

import bluesky.plan_stubs as bps


def request_input(msg="", default="n", agree="y", bypass=False):
    """
    Request input from the user.  Returns ``True`` if confirmed.

    Return whether (lower case) the response from user (or default) starts with the text
    of ``agree`` or ``bypass is True``.

    PARAMETERS

    msg str:
        Message text to be printed.
        (default: ``""``)
    default str:
        Default response if user accepts default.
        (default: ``"n"``)
    agree (str or list):
        User (or default) response must start with this text for ``True``. If a
        list of strings is provided, response must match (lower case) one of the
        strings in the list.
        (default: ``"y"``)
    bypass bool:
        Allow for automated plans to bypass this request in-place.
        (default: ``False``)

    New in release 1.6.6
    """
    match = False
    if not bypass:
        #  note: caller should write ``msg`` to show the value(s) that match ``agree``
        full_text = f"{msg} [{default}] "
        r = yield from bps.input_plan(full_text)
        print(f"Response: {r = }")
        if len(r) == 0:
            r = default
        r = r.lower()
        if isinstance(agree, str):
            match = r.startswith(agree)
        elif isinstance(agree, list):
            match = r in [str(l).lower() for l in agree]
        else:
            raise TypeError(f"Unhandled type: agree={agree}")
    return bypass or match



from ophyd import EpicsSignal

from ..positioner_soft_done import PVPositionerSoftDone
from ..positioner_soft_done import PVPositionerSoftDoneWithStop
from ...synApps.swait import UserCalcsDevice
from ...tests import common_attribute_quantities_test
from ...tests import IOC
from ...tests import short_delay_for_EPICS_IOC_database_processing
from ...utils import run_in_thread

import pytest
import random
import time

PV_PREFIX = f"{IOC}gp:"
delay_active = False


@pytest.fixture(scope="function")
def pos():
    """Test Positioner based on two analogout PVs."""
    # fmt: off
    pos = PVPositionerSoftDoneWithStop(
        PV_PREFIX, readback_pv="float1", setpoint_pv="float2", name="pos"
    )
    # fmt: on
    pos.wait_for_connection()
    yield pos


@pytest.fixture(scope="function")
def prec():
    "PV precision."
    prec = EpicsSignal(f"{PV_PREFIX}float1.PREC", name="prec")
    prec.wait_for_connection()
    yield prec


@pytest.fixture(scope="function")
def rbv():
    "Writable readback value for tests."
    rbv = EpicsSignal(f"{PV_PREFIX}float1", name="rbv")
    rbv.wait_for_connection()
    yield rbv


@pytest.fixture(scope="function")
def calcpos():
    """Test Positioner based on swait (userCalc) PV."""
    user = UserCalcsDevice(IOC, name="user")
    user.wait_for_connection()
    user.enable.put("Enable")

    swait = user.calc10
    swait.wait_for_connection()
    swait.reset()
    swait.calculation.put("A+(B-A)*C")  # incremental move towards the goal
    swait.channels.A.input_pv.put(swait.calculated_value.pvname)  # readback
    # channel B: setpoint
    swait.channels.C.input_value.put(0.9)  # move fraction
    swait.scanning_rate.put(".1 second")

    calcpos = PVPositionerSoftDoneWithStop(
        "",
        readback_pv=swait.calculated_value.pvname,
        setpoint_pv=swait.channels.B.input_value.pvname,
        name="calcpos",
    )
    calcpos.wait_for_connection()
    yield calcpos


@run_in_thread
def delayed_complete(positioner, readback, delay=1):
    "Time-delayed completion of positioner move."
    global delay_active

    delay_active = True
    time.sleep(delay)
    readback.put(positioner.setpoint.get())
    delay_active = False


@run_in_thread
def delayed_stop(positioner, delay=1):
    "Time-delayed stop of positioner."
    time.sleep(delay)
    positioner.stop()


class Tst_PVPos(PVPositionerSoftDone):
    "To pass the readback_pv and setpoint_pv kwargs."

    def __init__(self, prefix, *args, **kwargs):
        super().__init__(prefix, *args, readback_pv="r", setpoint_pv="p", **kwargs)


class Tst_PVPosWStop(PVPositionerSoftDoneWithStop):
    "To pass the readback_pv and setpoint_pv kwargs."

    def __init__(self, prefix, *args, **kwargs):
        super().__init__(prefix, *args, readback_pv="r", setpoint_pv="p", **kwargs)


@pytest.mark.parametrize(
    "device, pv, connect, attr, expected",
    [
        [Tst_PVPos, "", False, "read_attrs", 2],
        [Tst_PVPos, "", False, "configuration_attrs", 3],
        [Tst_PVPos, "", False, "component_names", 6],
        [Tst_PVPosWStop, "", False, "read_attrs", 2],
        [Tst_PVPosWStop, "", False, "configuration_attrs", 3],
        [Tst_PVPosWStop, "", False, "component_names", 6],
    ],
)
def test_attribute_quantities(device, pv, connect, attr, expected):
    """Verify the quantities of the different attributes."""
    common_attribute_quantities_test(device, pv, connect, attr, expected)


@pytest.mark.parametrize(
    "klass, rb, sp",
    [
        [PVPositionerSoftDone, None, None],
        [PVPositionerSoftDone, "", ""],
        [PVPositionerSoftDone, "test", "test"],
        [PVPositionerSoftDoneWithStop, None, None],
        [PVPositionerSoftDoneWithStop, "", ""],
        [PVPositionerSoftDoneWithStop, "test", "test"],
    ],
)
def test_same_sp_and_rb(klass, rb, sp):
    with pytest.raises(ValueError) as exc:
        klass("", readback_pv=rb, setpoint_pv=sp, name="pos")
    assert str(exc.value).endswith("must have different values")


@pytest.mark.parametrize(
    "device, has_inposition",
    [[PVPositionerSoftDone, False], [PVPositionerSoftDoneWithStop, True]],
)
def test_structure(device, has_inposition):
    "actual PVs not necessary for this test"
    pos = device("", readback_pv="r", setpoint_pv="v", name="pos")
    assert isinstance(pos, device)
    assert hasattr(pos, "inposition") is has_inposition
    assert pos.precision is None
    assert pos.prefix == ""
    assert pos.readback.pvname == "r"
    assert pos.report_dmov_changes.get() is False
    assert pos.setpoint.pvname == "v"
    assert pos.done.get() is True
    assert pos.done_value is True
    assert pos.target.get() == "None"
    assert pos.tolerance.get() == -1


def test_put_and_stop(rbv, prec, pos):
    short_delay_for_EPICS_IOC_database_processing()
    assert pos.tolerance.get() == -1
    assert pos.precision == prec.get()

    # ensure starting value at 0.0
    pos.setpoint.put(0)
    rbv.put(1)  # make the readback to different
    short_delay_for_EPICS_IOC_database_processing()
    assert not pos.inposition

    rbv.put(0)  # make the readback match
    short_delay_for_EPICS_IOC_database_processing()
    assert pos.position == 0.0

    assert pos.done.get() is True
    assert pos.done_value is True
    assert pos.inposition

    # change the setpoint
    pos.setpoint.put(1)
    short_delay_for_EPICS_IOC_database_processing()
    assert not pos.inposition

    # change the readback to match
    rbv.put(1)
    short_delay_for_EPICS_IOC_database_processing()
    assert pos.inposition

    # change the setpoint
    pos.setpoint.put(0)
    short_delay_for_EPICS_IOC_database_processing()
    assert not pos.inposition

    # move the readback part-way, but move is not over yet
    rbv.put(0.5)
    short_delay_for_EPICS_IOC_database_processing()
    assert not pos.inposition

    # force a stop now
    pos.stop()
    short_delay_for_EPICS_IOC_database_processing()
    pos.cb_readback()
    short_delay_for_EPICS_IOC_database_processing()
    assert pos.setpoint.get() == 0.5
    assert pos.readback.get() == 0.5
    assert pos.position == 0.5
    assert pos.inposition


def test_move_and_stop_nonzero(rbv, pos):
    short_delay_for_EPICS_IOC_database_processing()

    # move to non-zero
    longer_delay = 2
    target = round(2 + 5 * random.random(), 2)
    delayed_complete(pos, rbv, delay=longer_delay)
    t0 = time.time()  # time it
    status = pos.move(target)  # readback set by delayed_complete()
    dt = time.time() - t0
    assert status.done
    assert status.success
    assert status.elapsed >= longer_delay
    short_delay_for_EPICS_IOC_database_processing()
    assert dt >= longer_delay
    assert pos.inposition


def test_move_and_stopped_early(rbv, pos):
    short_delay_for_EPICS_IOC_database_processing()

    # first, move to some random, non-zero, initial position
    target = round(2 + 5 * random.random(), 2)
    delayed_complete(pos, rbv, delay=0.5)
    status = pos.move(target)  # readback set by delayed_complete()
    short_delay_for_EPICS_IOC_database_processing()
    assert status.done
    assert status.success
    # assert status.elapsed >= 0.5  # can't assure for initial move

    # move that is stopped before reaching the target
    longer_delay = 2
    t0 = time.time()  # time it
    delayed_stop(pos, longer_delay)
    assert pos.inposition
    status = pos.move(target - 1)  # readback set by delayed_stop()
    assert status.done
    assert status.success
    dt = time.time() - t0
    # assert status.elapsed >= longer_delay  # FIXME: fails sometimes
    # assert dt >= longer_delay  # TODO: restore
    assert dt >= 0

    short_delay_for_EPICS_IOC_database_processing()
    assert pos.setpoint.get() == target
    assert round(pos.position, 2) == target
    assert pos.inposition


def confirm_in_position(positioner):
    """Apply the 'inposition' property code."""
    positioner.get(use_monitor=False)
    reading = positioner.read()
    sp = reading[positioner.setpoint.name]["value"]
    rb = reading[positioner.readback.name]["value"]
    tol = positioner.actual_tolerance
    assert abs(rb - sp) <= tol, f"setpoint={sp}, readback={rb}, tolerance={tol}"


@pytest.mark.local
@pytest.mark.parametrize(
    "target",
    # fmt: off
    [-1] * 3
    + [0] * 5
    + [1, -1, -1, 1, 1, 2, 0, 1, -1, -1]
    + [round(2 + 5 * random.random(), 2), 0.1, -0.15, -1]
    + [0] * 5,
    # fmt: on
)
def test_target_practice(target, rbv, pos):
    """
    Watch for random errors that fail to reach intended target position.

    Test battery is a sequence of target positions, including repeated moves to
    the same location and pathologies involving numerical zero.
    """
    short_delay_for_EPICS_IOC_database_processing()

    def confirm_sequence_timings():
        """Confirm that values are set in this order."""
        reading = pos.read()
        t_rb = reading[pos.readback.name]["timestamp"]
        t_sp = reading[pos.setpoint.name]["timestamp"]
        t_rbv = rbv.read()[rbv.name]["timestamp"]

        assert t_sp <= t_rbv, f"rbv={t_rbv-t_sp}, readback input not updated"
        assert t_rbv <= t_rb, f"rb={t_rb-t_sp}, readback value not updated"

    # start from known position
    known_start_position = -1
    starting_rb_count = pos._rb_count
    starting_sp_count = pos._sp_count

    # initiate the move to the known position
    pos.setpoint.put(known_start_position)
    # short_delay_for_EPICS_IOC_database_processing()
    assert pos._sp_count > starting_sp_count, "cb_setpoint() was not called"

    # complete the initial move by updating the readback signal
    rbv.put(known_start_position)  # note: pos.readback is read-only
    pos.cb_readback()
    short_delay_for_EPICS_IOC_database_processing(1)
    # time.sleep(0.03)
    assert pos._rb_count > starting_rb_count, "cb_readback() was not called"

    # confirm all are ready
    # confirm_sequence_timings()
    confirm_in_position(pos)
    assert pos.setpoint.get(use_monitor=False) == known_start_position
    assert round(pos.readback.get(use_monitor=False), 2) == known_start_position
    assert pos.inposition, f"target={pos.setpoint.get()}, readback={pos.position}"

    # pick a random time to change RBV after the move starts
    rbv_delay = round(0.2 + 0.7 * random.random(), 1)

    # start code that will update the RBV later
    delayed_complete(pos, rbv, delay=rbv_delay)

    # start the move and wait for RBV to be updated
    diff = target - pos.setpoint.get()
    is_new_target = round(abs(diff), 2) > pos.actual_tolerance
    status = pos.move(target)

    short_delay_for_EPICS_IOC_database_processing()

    # assert not delay_active
    assert pos.setpoint.get(use_monitor=False) == target

    # note: pos.position has been failing (issue #668)
    # Replace ``pos.position`` and force a CA get from the IOC.
    assert round(pos.readback.get(use_monitor=False), 2) == target
    if is_new_target:
        confirm_sequence_timings()
    confirm_in_position(pos)
    assert pos.inposition

    assert status.done
    assert status.success


@pytest.mark.local
@pytest.mark.parametrize(
    "target",
    # fmt: off
    [-1] * 3
    + [0] * 5
    + [1, -1, -1, 1, 1, 2, 0, 1, -1, -1]
    + [round(2 + 5 * random.random(), 2), 0.1, -0.15, -1]
    + [0] * 5,
    # fmt: on
)
def test_move_calcpos(target, calcpos):
    """Demonstrate simpler test with positioner that updates its own RBV."""
    status = calcpos.move(target)
    assert status.elapsed > 0, str(status)
    confirm_in_position(calcpos)
    if not calcpos.inposition:
        calcpos.cb_readback()  # nudge it one more time
    assert calcpos.inposition, str(pos)

    time.sleep(0.2)  # pause between tests


@pytest.mark.local
@pytest.mark.parametrize(
    # fmt: off
    "target", [-55, -1.2345, -1, -1, -0.1, -0.1, 0, 0, 0, 0.1, 0.1, 1, 1, 1.2345, 55]
    # fmt: on
)
def test_same_position_725(target, calcpos):
    # Before anything is changed.
    confirm_in_position(calcpos)

    # Confirm the initial position is as expected.
    if calcpos.position == target:
        user = UserCalcsDevice(IOC, name="user")
        user.wait_for_connection()
        swait = user.calc10
        # First, move away from the target.
        away_target = -target or 0.1 * (1 + random.random())
        status = calcpos.move(away_target)
        assert status.done
        assert status.elapsed > 0, str(status)
        assert swait.channels.B.input_value.get() == away_target, f"{swait.channels.B.input_value.get()}  {away_target=}  {target=}"
        confirm_in_position(calcpos)

    # Move to the target position.
    status = calcpos.move(target)
    assert status.done
    assert status.elapsed > 0, str(status)
    confirm_in_position(calcpos)

    # Do it again (the reason for issue #725).
    status = calcpos.move(target)
    assert status.done
    assert status.elapsed > 0, str(status)
    confirm_in_position(calcpos)



"""
unit tests for the filewriters
"""

import databroker
import h5py
import os
import pytest
import spec2nexus.spec
import tempfile

from ..callback_base import FileWriterCallbackBase
from .. import NEXUS_FILE_EXTENSION
from .. import NEXUS_RELEASE
from .. import NXWriter
from .. import NXWriterAPS
from .. import SpecWriterCallback


CATALOG = "usaxs_test"
COUNT = "555a604"  # <-- uid,  scan_id: 2
TUNE_AR = 103  # <-- scan_id,  uid: "3554003"
TUNE_MR = 108  # <-- scan_id,  uid: "2ffe4d8"


@pytest.fixture(scope="function")
def cat():
    cat = databroker.catalog[CATALOG]
    return cat


@pytest.fixture(scope="function")
def tempdir():
    tempdir = tempfile.mkdtemp()
    return tempdir


def to_string(text):
    """Make string comparisons consistent."""
    if isinstance(text, bytes):
        text = text.decode()
    return text


def write_stream(specwriter, stream):
    """write the doc stream to the file"""
    for tag, doc in stream:
        specwriter.receiver(tag, doc)


def test_catalog(cat):
    assert len(cat) == 10
    assert cat["99f"].metadata["start"]["scan_id"] == 2
    assert cat[110].metadata["start"]["uid"][:5] == "19965"


def test_replay(cat):
    run = cat.v1[COUNT]
    assert len(list(run.documents())) == 7
    di = iter(run.documents())
    key, doc = next(di)
    assert key == "start"
    key, doc = next(di)
    assert key == "descriptor"
    key, doc = next(di)
    assert key == "descriptor"
    key, doc = next(di)
    assert key == "event"
    key, doc = next(di)
    assert key == "event"
    key, doc = next(di)
    assert key == "event"
    key, doc = next(di)
    assert key == "stop"


def test_FileWriterCallbackBase(cat, capsys):
    callback = FileWriterCallbackBase()
    for uid in cat:
        run = cat.v1[uid]
        callback.clear()
        assert callback.uid is None
        assert callback.scanning is False

        for tag, doc in run.documents():
            callback.receiver(tag, doc)
            out, err = capsys.readouterr()
            if tag == "start":
                assert callback.scanning is True
            elif tag == "stop":
                # callback.writer called after stop document
                lines = out.splitlines()
                assert len(lines) > 0
                assert lines[0] == "print to console"
                assert lines[1].startswith("suggested file name: ")

        assert callback.plan_name == run["start"]["plan_name"]
        assert not callback.scanning
        assert callback.uid is not None
        assert callback.exit_status is not None
        assert callback.start_time is not None
        assert callback.stop_time is not None
        assert callback.stop_reason is not None
        assert callback.stop_time > callback.start_time
        assert len(callback.acquisitions) == len(callback.streams)
        assert callback.scan_id > 0


def test_NXWriterAPS(cat, tempdir):
    callback = NXWriterAPS()
    callback.file_path = tempdir

    run = cat.v1[TUNE_MR]
    for tag, doc in run.documents():
        callback.receiver(tag, doc)

    fname = callback.make_file_name()
    assert os.path.exists(fname)
    with h5py.File(fname, "r") as nxroot:
        assert "/entry/instrument/source" in nxroot
        nxsource = nxroot["/entry/instrument/source"]
        assert to_string(nxsource["name"][()]) == "Advanced Photon Source"
        assert nxsource["name"].attrs["short_name"] == "APS"
        assert to_string(nxsource["type"][()]) == "Synchrotron X-ray Source"
        assert to_string(nxsource["probe"][()]) == "x-ray"
        assert nxsource["energy"][()] == 6
        assert nxsource["energy"].attrs["units"] == "GeV"
        assert "/entry/instrument/undulator" in nxroot


def test_NXWriter_default_plot(cat, tempdir):
    callback = NXWriterAPS()
    callback.file_path = tempdir

    run = cat.v1[TUNE_MR]
    for tag, doc in run.documents():
        callback.receiver(tag, doc)

    fname = callback.make_file_name()
    assert os.path.exists(fname)
    with h5py.File(fname, "r") as nxroot:
        assert nxroot.attrs["default"] == "entry"
        assert nxroot["/entry"].attrs["default"] == "data"
        nxdata = nxroot["/entry/data"]
        signal = nxdata.attrs.get("signal")
        assert signal is not None
        assert signal in nxdata
        axes = nxdata.attrs.get("axes")
        assert axes is not None
        assert len(axes) == 1
        assert axes[0] in nxdata
        assert axes[0] != signal


def test_NXWriter_make_file_name(tempdir):
    import datetime
    import dateutil.tz

    callback = NXWriter()

    assert callback.file_path is None
    assert callback.scan_id is None
    assert callback.start_time is None
    assert callback.uid is None

    with pytest.raises(TypeError) as exinfo:
        # error from f"-S{self.scan_id:05d}"
        callback.make_file_name()
    assert exinfo is not None
    assert "NoneType" in str(exinfo.value), f"{callback=} {exinfo=}"
    assert "an integer" in str(exinfo.value), f"{callback=} {exinfo=}"

    callback.start_time = 1000  # 1969-12-31T18:16:40
    with pytest.raises(TypeError) as exinfo:
        # error from datetime.datetime.fromtimestamp
        callback.make_file_name()
    expected = "unsupported format string passed to NoneType.__format__"
    assert expected == str(exinfo.value)

    callback.scan_id = 9876
    with pytest.raises(TypeError) as exinfo:
        callback.make_file_name()
    assert "'NoneType' object is not subscriptable" == str(exinfo.value)

    callback.uid = "012345678901234567890123456789"
    fname = callback.make_file_name()
    # https://github.com/BCDA-APS/apstools/issues/345
    tz_aps = dateutil.tz.gettz("America/Chicago")
    tz_local = dateutil.tz.tzlocal()
    # datetime at the APS: 1969-12-31-18:16:40"
    dt = datetime.datetime.fromtimestamp(callback.start_time, tz_aps)
    # datetime on the local python test workstation
    dt_local = dt.astimezone(tz_local)
    # render `expected` in local time zone
    expected = dt_local.strftime("%Y%m%d-%H%M%S")
    expected += f"-S{9876:05d}"
    expected += "-0123456"
    expected += f".{NEXUS_FILE_EXTENSION}"
    assert os.path.split(fname)[-1] == expected
    assert os.path.dirname(fname) == os.getcwd()

    callback.file_path = tempdir
    fname = callback.make_file_name()
    assert os.path.dirname(fname) == tempdir


def test_NXWriter_receiver_battery(cat, tempdir):
    callback = NXWriter()
    assert callback.nexus_release == NEXUS_RELEASE
    assert callback.file_extension == NEXUS_FILE_EXTENSION

    for uid in cat:
        run = cat.v1[uid]
        callback.clear()
        callback.file_path = tempdir
        assert callback.uid is None
        assert not callback.scanning

        for tag, doc in run.documents():
            callback.receiver(tag, doc)

        fname = callback.make_file_name()
        assert os.path.exists(fname)
        with h5py.File(fname, "r") as nxroot:
            assert nxroot.attrs["NeXus_version"] == NEXUS_RELEASE
            assert nxroot.attrs["creator"] == callback.__class__.__name__

            assert "/entry" in nxroot
            nxentry = nxroot["/entry"]
            assert to_string(nxentry["entry_identifier"][()]) == to_string(callback.uid)
            assert to_string(nxentry["plan_name"][()]) == to_string(callback.plan_name)

            assert "instrument/bluesky" in nxentry
            bluesky_group = nxentry["instrument/bluesky"]
            assert "metadata" in bluesky_group
            assert "streams" in bluesky_group
            assert len(bluesky_group["streams"]) == len(callback.streams)

            assert "/entry/instrument/undulator" not in nxroot


def test_SpecWriterCallback_writer_default_name(cat, tempdir):
    specwriter = SpecWriterCallback()
    path = os.path.abspath(os.path.dirname(specwriter.spec_filename))
    assert path != tempdir  # "default file not in tempdir"
    assert path == os.path.abspath(os.getcwd())  # "default file to go in pwd"

    # change the directory
    specwriter.spec_filename = os.path.join(
        tempdir, specwriter.spec_filename
    )

    assert not os.path.exists(specwriter.spec_filename)  # "data file not created yet"
    write_stream(specwriter, cat.v1[TUNE_MR].documents())
    assert os.path.exists(specwriter.spec_filename)  # "data file created"

    sdf = spec2nexus.spec.SpecDataFile(specwriter.spec_filename)
    assert len(sdf.headers) == 1
    assert len(sdf.scans) == 1

    # check that the #N line is written properly (issue #203)
    scans = sdf.getScanNumbers()
    assert 108 not in scans
    assert str(108) in scans
    scan = sdf.getScan(108)
    assert scan.N[0] == len(scan.L)

    assert scan.header.raw.find("\n#O0 \n") > 0
    assert scan.header.raw.find("\n#o0 \n") > 0
    assert len(scan.header.O) == 1
    assert len(scan.header.O[0]) == 0
    assert len(scan.header.o) == 1
    assert len(scan.header.o[0]) == 0

    assert scan.raw.find("\n#P0 \n") > 0
    assert len(scan.P) == 1
    assert len(scan.P[0]) == 0
    assert len(scan.positioner) == 0


def test_SpecWriterCallback_writer_filename(cat, tempdir):
    testfile = os.path.join(tempdir, "tune_mr.dat")
    if os.path.exists(testfile):
        os.remove(testfile)
    specwriter = SpecWriterCallback(filename=testfile)

    assert isinstance(specwriter, SpecWriterCallback)
    assert specwriter.spec_filename == testfile

    assert not os.path.exists(testfile)
    write_stream(specwriter, cat.v1[TUNE_MR].documents())
    assert os.path.exists(testfile)


def test_SpecWriterCallback_newfile_exists(cat, tempdir):
    testfile = os.path.join(tempdir, "tune_mr.dat")
    if os.path.exists(testfile):
        os.remove(testfile)
    specwriter = SpecWriterCallback(
        filename=testfile
    )

    from .. import SCAN_ID_RESET_VALUE

    assert SCAN_ID_RESET_VALUE == 0  # "default reset scan id"

    write_stream(specwriter, cat.v1[TUNE_MR].documents())
    assert os.path.exists(testfile)  # "data file created"

    raised = False
    try:
        specwriter.newfile(filename=testfile)
    except ValueError:
        raised = True
    finally:
        assert not raised  # "file exists"
    assert specwriter.reset_scan_id == 0  # "check scan id"

    class my_RunEngine:
        # duck type model *strictly* for testing *here*
        md = dict(scan_id=SCAN_ID_RESET_VALUE)

    RE = my_RunEngine()

    specwriter.scan_id = -5  # an unusual value for testing only
    RE.md["scan_id"] = -10  # an unusual value for testing only
    specwriter.newfile(filename=testfile, scan_id=None, RE=RE)
    assert specwriter.scan_id == 108  # "scan_id unchanged"
    assert RE.md["scan_id"] == 108  # "RE.md['scan_id'] unchanged"

    specwriter.scan_id = -5  # an unusual value for testing only
    RE.md["scan_id"] = -10  # an unusual value for testing only
    specwriter.newfile(filename=testfile, scan_id=False, RE=RE)
    assert specwriter.scan_id == 108  # "scan_id unchanged"
    assert RE.md["scan_id"] == 108  # "RE.md['scan_id'] unchanged"

    specwriter.scan_id = -5  # an unusual value for testing only
    RE.md["scan_id"] = -10  # an unusual value for testing only
    specwriter.newfile(filename=testfile, scan_id=True, RE=RE)
    assert specwriter.scan_id == 108  # "scan_id reset"
    assert RE.md["scan_id"] == 108  # "RE.md['scan_id'] reset"

    for n, s in {"0": 108, "108": 108, "110": 110}.items():
        specwriter.scan_id = -5  # an unusual value for testing only
        RE.md["scan_id"] = -10  # an unusual value for testing only
        specwriter.newfile(filename=testfile, scan_id=int(n), RE=RE)
        assert specwriter.scan_id == s
        assert RE.md["scan_id"] == s


def test_SpecWriterCallback__rebuild_scan_command(cat):
    from ..spec_file_writer import _rebuild_scan_command

    start_docs = []
    for tag, doc in cat.v1[TUNE_MR].documents():
        if tag == "start":
            start_docs.append(doc)
    assert len(start_docs) == 1  # "unique start doc found"

    doc = start_docs[0]
    expected = "108  tune_mr()"
    result = _rebuild_scan_command(doc)
    assert result == expected  # "rebuilt #S line"


def test_SpecWriterCallback_spec_comment(cat, tempdir):
    from .. import spec_comment

    # spec_comment(comment, doc=None, writer=None)
    testfile = os.path.join(tempdir, "spec_comment.dat")
    if os.path.exists(testfile):
        os.remove(testfile)
    specwriter = SpecWriterCallback(
        filename=testfile
    )

    for category in "buffered_comments comments".split():
        for k in "start stop descriptor event".split():
            o = getattr(specwriter, category)
            assert len(o[k]) == 0

    # insert comments with every document
    spec_comment(
        "TESTING: Should appear within start doc",
        doc=None,
        writer=specwriter,
    )

    for idx, document in enumerate(cat.v1[TUNE_MR].documents()):
        tag, doc = document
        msg = f"TESTING: document {idx+1}: '{tag}' %s specwriter.receiver"
        spec_comment(msg % "before", doc=tag, writer=specwriter)
        specwriter.receiver(tag, doc)
        if tag == "stop":
            # since stop doc was received, this appears in the next scan
            spec_comment(
                str(msg % "before") + " (appears at END of next scan)",
                doc=tag,
                writer=specwriter,
            )
        else:
            spec_comment(msg % "after", doc=tag, writer=specwriter)

    assert len(specwriter.buffered_comments["stop"]) == 1

    # since stop doc was received, this appears in the next scan
    spec_comment(
        "TESTING: Appears at END of next scan",
        doc="stop",
        writer=specwriter,
    )

    assert len(specwriter.buffered_comments["stop"]) == 2
    write_stream(specwriter, cat.v1[TUNE_AR].documents())

    for k in "start descriptor event".split():
        o = specwriter.buffered_comments
        assert len(o[k]) == 0
    expected = dict(start=2, stop=5, event=0, descriptor=0)
    for k, v in expected.items():
        assert len(specwriter.comments[k]) == v



"""
Support for ``stage_sigs``
+++++++++++++++++++++++++++++++++++++++

Often when writing a plan, it is desired to change the staging of one or more
``ophyd.Devices`` by modifying the ``device.stage_sigs`` dictionary(ies), then
reset them to original values after the plan.

.. autosummary::

   ~restorable_stage_sigs
   ~stage_sigs_wrapper

EXAMPLE:

For example, move the motor at 2 units/second and count the scaler for 0.5
seconds during a custom plan.  After the plan, both scaler and motor will have
their original ``.stage_sigs`` dictionaries.

.. code-block:: python
    :linenos:

    @restorable_stage_sigs([scaler, motor])
    def my_plan(start, finish, npts, ct=1, v=1):
        scaler.stage_sigs = dict(preset_time=ct)
        motor.stage_sigs = dict(velocity=v)
        yield from bp.rel_scan([scaler], motor, start, finish, npts)

    RE(my_plan(-1, 1, 11, ct=0.5, v=2))

(new in apstools release 1.6.9)
"""

from bluesky.utils import make_decorator
import logging


logger = logging.getLogger(__name__)


def stage_sigs_wrapper(user_plan, devices):
    """
    Save stage_sigs from each device and restore after the user_plan.

    The user_plan is free to modify the stage_sigs of each device
    without further need to preserve original values.
    """

    def display(preface):
        for device in devices:
            logger.debug("%s: %s.stage_sigs: %s", preface, device.name, device.stage_sigs)

    def _restore():
        for device in reversed(devices):
            device.stage_sigs = original[device].copy()
        display("AFTER restore")

    original = {}
    display("ORIGINAL")
    for device in devices:
        original[device] = device.stage_sigs.copy()

    try:
        display("BEFORE plan")
        yield from user_plan
        display("AFTER plan")
    finally:
        _restore()


restorable_stage_sigs = make_decorator(stage_sigs_wrapper)



"""
XIA PF4 Filters
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~Pf4FilterSingle
   ~Pf4FilterDual
   ~Pf4FilterTriple
   ~Pf4FilterCommon
   ~Pf4FilterBank
   ~DualPf4FilterBox
"""

from ophyd import Component
from ophyd import Device
from ophyd import EpicsSignal
from ophyd import EpicsSignalRO
from ophyd import FormattedComponent


class Pf4FilterBank(Device):
    """
    A single module of XIA PF4 filters (4-blades).

    .. index:: Ophyd Device; Pf4FilterBank

    EXAMPLES::


        pf4B = Pf4FilterBank("ioc:pf4:", name="pf4B", bank="B")

        # -or-

        class MyTriplePf4(Pf4FilterCommon):
            A = Component(Pf4FilterBank, "", bank="A")
            B = Component(Pf4FilterBank, "", bank="B")
            C = Component(Pf4FilterBank, "", bank="C")

        pf4 = MyTriplePf4("ioc:pf4:", name="pf4")

    :see: https://github.com/epics-modules/optics/blob/master/opticsApp/Db/pf4bank.db
    """

    fPos = FormattedComponent(EpicsSignal, "{prefix}fPos{_bank}", kind="config")
    control = FormattedComponent(
        EpicsSignalRO, "{prefix}bank{_bank}", string=True, kind="config"
    )
    bits = FormattedComponent(EpicsSignalRO, "{prefix}bitFlag{_bank}", kind="config")
    transmission = FormattedComponent(
        EpicsSignalRO, "{prefix}trans{_bank}", kind="config"
    )

    def __init__(self, prefix, bank=None, **kwargs):
        self._bank = bank
        super().__init__(prefix, **kwargs)


class Pf4FilterCommon(Device):
    """
    XIA PF4 filters - common support.

    .. index:: Ophyd Device; Pf4FilterCommon

    Use ``Pf4FilterCommon`` to build support for a configuration
    of PF4 filters (such as 3 or 4 filter banks).

    EXAMPLE::

        class MyTriplePf4(Pf4FilterCommon):
            A = Component(Pf4FilterBank, "", bank="A")
            B = Component(Pf4FilterBank, "", bank="B")
            C = Component(Pf4FilterBank, "", bank="C")

        pf4 = MyTriplePf4("ioc:pf4:", name="pf4")

    :see: https://github.com/epics-modules/optics/blob/master/opticsApp/Db/pf4common.db
    """

    transmission = Component(EpicsSignalRO, "trans", kind="hinted")
    inverse_transmission = Component(EpicsSignalRO, "invTrans", kind="normal")

    thickness_Al_mm = Component(EpicsSignalRO, "filterAl", kind="config")
    thickness_Ti_mm = Component(EpicsSignalRO, "filterTi", kind="config")
    thickness_glass_mm = Component(EpicsSignalRO, "filterGlass", kind="config")

    energy_keV_local = Component(EpicsSignal, "E:local", kind="config")
    energy_keV_mono = Component(EpicsSignal, "displayEnergy", kind="config")

    mode = Component(EpicsSignal, "useMono", string=True, kind="config")


class Pf4FilterSingle(Pf4FilterCommon):
    """XIA PF4 Filter: one set of 4 filters (A)."""

    A = Component(Pf4FilterBank, "", bank="A")


class Pf4FilterDual(Pf4FilterSingle):
    """XIA PF4 Filter: two sets of 4 filters (A, B)."""

    B = Component(Pf4FilterBank, "", bank="B")


class Pf4FilterTriple(Pf4FilterDual):
    """XIA PF4 Filter: three sets of 4 filters (A, B, C)."""

    C = Component(Pf4FilterBank, "", bank="C")


class DualPf4FilterBox(Device):
    """
    LEGACY (use Pf4FilterDual now): Dual (Al, Ti) Xia PF4 filter boxes

    Support from synApps (using Al, Ti foils)

    .. index:: Ophyd Device; DualPf4FilterBox

    EXAMPLE::

        pf4 = DualPf4FilterBox("2bmb:pf4:", name="pf4")
        pf4_AlTi = DualPf4FilterBox("9idcRIO:pf4:", name="pf4_AlTi")

    """

    fPosA = Component(EpicsSignal, "fPosA")
    fPosB = Component(EpicsSignal, "fPosB")
    bankA = Component(EpicsSignalRO, "bankA")
    bankB = Component(EpicsSignalRO, "bankB")
    bitFlagA = Component(EpicsSignalRO, "bitFlagA")
    bitFlagB = Component(EpicsSignalRO, "bitFlagB")
    transmission = Component(EpicsSignalRO, "trans")
    transmission_a = Component(EpicsSignalRO, "transA")
    transmission_b = Component(EpicsSignalRO, "transB")
    inverse_transmission = Component(EpicsSignalRO, "invTrans")
    thickness_Al_mm = Component(EpicsSignalRO, "filterAl")
    thickness_Ti_mm = Component(EpicsSignalRO, "filterTi")
    thickness_glass_mm = Component(EpicsSignalRO, "filterGlass")
    energy_keV_local = Component(EpicsSignal, "E:local")
    energy_keV_mono = Component(EpicsSignal, "displayEnergy")
    mode = Component(EpicsSignal, "useMono", string=True)

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
test issue #440: specwriter
"""

import intake
import numpy as np
import os
import zipfile

from .. import SpecWriterCallback
from ..spec_file_writer import _rebuild_scan_command


DATA_ARCHIVE = "440_specwriter_problem_run.zip"

PATH = os.path.dirname(__file__)
FULL_ZIP_FILE = os.path.join(PATH, DATA_ARCHIVE)

TMP_CATALOG = os.path.join(
    "/tmp", DATA_ARCHIVE.split(".")[0], "catalog.yml"
)


def test_setup_comes_first():
    assert os.path.exists(FULL_ZIP_FILE)

    with zipfile.ZipFile(FULL_ZIP_FILE, "r") as zip_ref:
        zip_ref.extractall("/tmp")

    assert os.path.exists(TMP_CATALOG)


def test_confirm_run_exists():
    assert os.path.exists(TMP_CATALOG)

    cat = intake.open_catalog(TMP_CATALOG)
    assert "packed_catalog" in cat

    cat = cat["packed_catalog"]
    assert len(cat) == 1
    assert "624e776a-a914-4a74-8841-babf1591fb29" in cat


def test_specwriter():
    # The problem does not appear when using data from the databroker.
    # Verify that is the case now.
    os.chdir("/tmp")
    specfile = "issue240.spec"
    if os.path.exists(specfile):
        os.remove(specfile)
    specwriter = SpecWriterCallback()
    specwriter.newfile(specfile)
    db = intake.open_catalog(TMP_CATALOG)["packed_catalog"].v1
    h = db[-1]
    for key, doc in db.get_documents(h):
        specwriter.receiver(key, doc)
        assert "relative_energy" not in doc
    assert os.path.exists(specwriter.spec_filename)

    with open(specwriter.spec_filename, "r") as f:
        line = ""
        while not line.startswith("#S 287"):
            line = f.readline()
        assert line.endswith("-0.05]})\n")

        # Next line should start with "#D "
        # The reported error had numbers from previous line (wrapped)
        line = f.readline()
        assert line.startswith("#D ")

    # The problem comes up if one of the arguments is a numpy.array.
    # So we must replay the document stream and modify the right
    # structure as it passes by.
    # This structure is in the start document, which is first.
    # Note: we don't have to write the whole SPEC file again,
    # just test if _rebuild_scan_command(start_doc) is one line.

    hh = db.get_documents(h)
    key, doc = next(hh)
    arr = doc["plan_args"]["qx_setup"]["relative_energy"]
    assert isinstance(arr, list)
    assert len(arr) == 183

    cmd = _rebuild_scan_command(doc)
    assert len(cmd.strip().splitlines()) == 1

    # Now, make it a numpy array and test again.
    arr = np.array(arr)
    assert isinstance(arr, np.ndarray)
    # modify the start doc
    doc["plan_args"]["qx_setup"]["relative_energy"] = arr
    cmd = _rebuild_scan_command(doc)  # FIXME: <-----
    assert len(cmd.strip().splitlines()) == 1



"""
Base class for Device Mixins
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~DeviceMixinBase
"""

from ophyd import Device


class DeviceMixinBase(Device):
    """
    Base class for apstools Device mixin classes

    .. index:: Ophyd Device Mixin; DeviceMixinBase
    """

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""Test the alignment plans."""

# TODO: TuneAxis
# TODO: tune_axes
# TODO: TuneResults

from .. import alignment
from ... import utils
from ...devices import SynPseudoVoigt
from ...synApps import setup_lorentzian_swait
from ...synApps import SwaitRecord
from ...tests import IOC
from bluesky import RunEngine
from bluesky.callbacks import best_effort
import bluesky.plans as bp
import bluesky.plan_stubs as bps
import databroker
import numpy
import ophyd
import pytest

bec = best_effort.BestEffortCallback()
cat = databroker.temp()
RE = RunEngine({})
RE.subscribe(cat.v1.insert)
RE.subscribe(bec)
bec.enable_plots()
# utils.ipython_shell_namespace()["bec"] = bec
# globals()["bec"] = bec

axis = ophyd.EpicsSignal(f"{IOC}gp:float1", name="axis")
m1 = ophyd.EpicsMotor(f"{IOC}m1", name="m1")
noisy = ophyd.EpicsSignalRO(f"{IOC}userCalc1.VAL", name="noisy")
scaler1 = ophyd.scaler.ScalerCH(f"{IOC}scaler1", name="scaler1")
swait = SwaitRecord(f"{IOC}userCalc1", name="swait")

for obj in (axis, m1, noisy, scaler1, swait):
    obj.wait_for_connection()

scaler1.select_channels()

# First, must be connected to m1.
pvoigt = SynPseudoVoigt(name="pvoigt", motor=axis, motor_field=axis.name)
pvoigt.kind = "hinted"


def change_noisy_parameters(fwhm=0.15, peak=10000, noise=0.08):
    """
    Setup the swait record with new random numbers.
    BLOCKING calls, not a bluesky plan
    """
    swait.reset()
    setup_lorentzian_swait(
        swait,
        m1.user_readback,
        center=2 * numpy.random.random() - 1,
        width=fwhm * numpy.random.random(),
        scale=peak * (9 + numpy.random.random()),
        noise=noise * (0.01 + numpy.random.random()),
    )


def get_position(obj, digits=3):
    if isinstance(obj, ophyd.EpicsMotor):
        position = obj.position
    else:
        position = obj.get()
    return round(position, digits)


@pytest.mark.parametrize("obj", [axis, m1, noisy, pvoigt, scaler1, swait])
def test_connected(obj):
    assert obj.connected


def test_SynPseudoVoigt_randomize():
    signal = SynPseudoVoigt("signal", axis, axis.name)
    # check default settings
    assert signal.center == 0
    assert signal.eta == 0.5
    assert signal.scale == 1
    assert signal.sigma == 1
    assert signal.bkg == 0
    assert signal.noise is None
    assert signal.noise_multiplier == 1

    signal.randomize_parameters()
    # confirm non-default settings
    assert -1 <= signal.center <= 1
    assert 0.2 <= signal.eta <= 0.7
    assert 95_000 <= signal.scale <= 105_000
    assert 0.001 <= signal.sigma <= 0.051
    assert 0 <= signal.bkg <= 0.01
    assert signal.noise == "poisson"
    assert signal.noise_multiplier == 1


@pytest.mark.parametrize(
    "signal, mover, start, finish, npts, feature",
    [
        # FIXME: [noisy, m1, -1.2, 1.2, 11, "max"],  # slower
        [pvoigt, axis, -1.2, 1.2, 11, "max"],  # faster
        [pvoigt, axis, -1.2, 1.2, 11, "cen"],
        [pvoigt, axis, -1.2, 1.2, 51, "com"],
    ],
)
def test_direct_implementation_with_rel_scan(signal, mover, start, finish, npts, feature):
    RE(bps.mv(mover, 0))
    assert get_position(mover) == 0.0

    if isinstance(signal, SynPseudoVoigt):
        signal.randomize_parameters(scale=250_000)
    # FIXME: else:
    #     change_noisy_parameters()

    RE(bp.rel_scan([signal], mover, start, finish, npts))
    analysis = bec.peaks[feature][signal.name]
    if isinstance(analysis, tuple):
        analysis = analysis[0]
    RE(bps.mv(mover, analysis))
    assert get_position(mover) == round(analysis, 3)

    if feature in "max cen".split():
        # Test if ended close to the expected value.
        if isinstance(signal, SynPseudoVoigt):
            center = signal.center
            width = signal.sigma * 2.355  # FWHM approx.
        # FIXME: else:
        #     center = swait.channels.B.input_value.get()
        #     width = swait.channels.C.input_value.get()
        position = get_position(mover)
        lo = center-width
        hi = center+width
        fwhm = bec.peaks["fwhm"][signal.name]
        assert lo <= position+fwhm, f"{lo=} {position=} {fwhm=}"
        assert position-fwhm < hi, f"{position=} {fwhm=} {hi=}"


@pytest.mark.parametrize(
    "signals, mover, start, finish, npts, feature, rescan",
    [
        [noisy, m1, -1.2, 1.2, 11, "max", True],  # slower, is motor
        [pvoigt, axis, -1.2, 1.2, 11, "cen", True],  # faster, is ao (float)
        [pvoigt, axis, -1.2, 1.2, 11, "max", True],
        [[pvoigt], axis, -1.2, 1.2, 11, "max", True],  # list
        [[pvoigt, noisy, scaler1], axis, -1.2, 1.2, 11, "max", True],  # more than one detector
        [(pvoigt), axis, -1.2, 1.2, 11, "max", True],  # tuple
        [pvoigt, axis, -1.2, 1.2, 11, "cen", False],
        [pvoigt, axis, -1.2, 1.2, 11, "com", False],
        [pvoigt, axis, -1.2, 1.2, 11, "max", False],
        [pvoigt, axis, -1.2, 1.2, 11, "min", False],  # pathological
    ],
)
def test_lineup(signals, mover, start, finish, npts, feature, rescan):
    if isinstance(signals, SynPseudoVoigt):
        signals.randomize_parameters(scale=250_000, bkg=0.000_000_000_1)
    else:
        change_noisy_parameters()

    RE(bps.mv(mover, 0))
    assert get_position(mover) == 0.0

    RE(alignment.lineup(signals, mover, start, finish, npts, feature=feature, rescan=rescan, bec=bec))

    # if rescan and feature in "max cen".split():
    #     # Test if mover position is within width of center.
    #     if isinstance(signals, SynPseudoVoigt):
    #         center = signals.center
    #         width = signals.sigma * 2.355  # FWHM approx.
    #     else:  # FIXME: multiple signals
    #         center = swait.channels.B.input_value.get()
    #         width = swait.channels.C.input_value.get()
    #     # FIXME: fails to find the bec analysis in lineup()
    #     position = get_position(mover)
    #     lo = center-width
    #     hi = center+width
    #     # assert lo <= position <= hi, f"{bec=} {bec.peaks=} {position=} {center=} {width=}"



"""
Statistical peak analysis functions
+++++++++++++++++++++++++++++++++++++++

Uses *pysumreg* package (https://prjemian.github.io/pysumreg/) to obtain summary
statistics.

.. autosummary::

   ~analyze_1D
   ~analyze_2D
"""

__all__ = """
    analyze_1D
    analyze_2D
""".split()

from pysumreg import SummationRegisters
import logging
import numpy as np

logger = logging.getLogger(__name__)
logger.info(__file__)


def analyze_1D(y_arr, x_arr=None):
    """
    Measures of 1D data peak center & width.

    Return result is a dictionary prepared by the
    ``to_dict(use_registers=True)`` method of the
    ``pysumreg.SummationRegisters()`` class.

    Example::

        {'mean_x': 2.0, 'mean_y': 7.2,
        'stddev_x': 1.5811388300841898, 'stddev_y': 3.3466401061363027,
        'slope': 0.0, 'intercept': 7.2, 'correlation': 0.0,
        'centroid': 2.0, 'sigma': 1.1547005383792515,
        'min_x': 1, 'max_x': 4, 'min_y': 4, 'max_y': 12,
        'x_at_max_y': 2,
        'n': 5, 'X': 10, 'Y': 36, 'XX': 30, 'XY': 72,
        'XXY': 192, 'YY': 304}
    """
    if x_arr is None:
        x_arr = list(range(len(y_arr)))
    if len(x_arr) != len(y_arr):
        raise ValueError("x and y arrays are not of the same length.")

    regs = SummationRegisters()
    regs.clear()
    for u, v in zip(x_arr, y_arr):
        regs.add(u, v)

    return regs.to_dict(use_registers=True)


def analyze_2D(image):
    """
    Analyze 2-D (image) data.

    Return result is a dictionary with the statistical results for a peak
    analysis, grouped in pairs (row, column) as it makes sense given
    ``frame[rows][columns]``.
    The :math:`x` values are the index number along the respective axis.

    For this image data::

        [
            [0, 1, 2, 1, 0],
            [1, 2, 3, 2, 1],
            [2, 3, 4, 10, 2],
            [1, 2, 3, 2, 1],
        ]

    This is the analysis::

        {'n': (5, 4),
         'centroid': (2.1628, 1.814),
         'sigma': (1.1192, 0.8695),
         'peak_position': (3, 2),
         'max_y': 10}
    """
    if not isinstance(image, np.ndarray):
        image = np.array(image)

    axis_0 = analyze_1D(image.sum(axis=0))
    axis_1 = analyze_1D(image.sum(axis=1))
    key_list = "n centroid sigma".split()
    results = {k: (axis_0[k], axis_1[k]) for k in key_list}
    k = "x_at_max_y"
    results["peak_position"] = (axis_0[k], axis_1[k])
    results["max_y"] = image[axis_1[k]][axis_0[k]]
    return results



import pytest
import time

from ..calcout import setup_incrementer_calcout
from ..calcout import CalcoutRecord
from ..calcout import UserCalcoutDevice
from ...tests import common_attribute_quantities_test
from ...tests import IOC
from ...tests import short_delay_for_EPICS_IOC_database_processing


@pytest.mark.parametrize(
    "device, pv, connect, attr, expected",
    [
        [CalcoutRecord, f"{IOC}userCalcOut10", False, "read_attrs", 12],
        [CalcoutRecord, f"{IOC}userCalcOut10", False, "configuration_attrs", 100],
        [CalcoutRecord, f"{IOC}userCalcOut10", True, "read()", 2],
        [CalcoutRecord, f"{IOC}userCalcOut10", True, "summary()", 220],

        [UserCalcoutDevice, IOC, False, "read_attrs", 130],
        [UserCalcoutDevice, IOC, False, "configuration_attrs", 1020],
        [UserCalcoutDevice, IOC, True, "read()", 20],
        [UserCalcoutDevice, IOC, True, "summary()", 2065],
    ]
)
def test_attribute_quantities(device, pv, connect, attr, expected):
    """Verify the quantities of the different attributes."""
    common_attribute_quantities_test(device, pv, connect, attr, expected)


def test_calcout_reset():
    user = UserCalcoutDevice(IOC, name="user")
    user.wait_for_connection()
    user.enable.put("Enable")

    calcout = user.calcout10
    assert isinstance(calcout, CalcoutRecord)
    calcout.enable.put("E")  # Note: only "E"

    setup_incrementer_calcout(calcout)
    time.sleep(0.2)
    assert calcout.description.get() == "incrementer"
    assert calcout.calculation.get() == "(A+1) % B"
    v1 = calcout.calculated_value.get()
    time.sleep(0.2)
    assert v1 < calcout.calculated_value.get()

    calcout.reset()
    short_delay_for_EPICS_IOC_database_processing()
    assert calcout.description.get() == calcout.prefix
    assert calcout.calculation.get() == "0"
    v1 = calcout.calculated_value.get()
    time.sleep(0.2)
    assert v1 == calcout.calculated_value.get()



from ..image_analysis import analyze_1D
from ..image_analysis import analyze_2D
from pysumreg import SummationRegisters
import pytest


@pytest.mark.parametrize(
    "data, expected, ndigits",
    [
        [
            # fmt: off
            dict(
                x=[.1, .2, .3, .4, .5, .6, .7, .8, .9, 1.0, 1.1],
                y=[0, 1, 2, 3, 4, 5, 4, 3, 2, 1, 0]
            ),
            # fmt: on
            dict(
                n=11,
                centroid=0.6,
                sigma=0.2,
                max_y=5,
                min_y=0,
                min_x=0.1,
                max_x=1.1,
                x_at_max_y=0.6,
                mean_x=0.6,
                mean_y=2.2727,
                stddev_x=0.33166,
                stddev_y=1.6787,
                correlation=0,
                intercept=2.2727,
                slope=0,
            ),
            4,
        ],
        [
            # fmt: off
            dict(
                y=[0, 1, 2, 3, 4, 5, 4, 3, 2, 1, 0]
            ),
            # fmt: on
            dict(
                n=11,
                mean_x=5.0,
                mean_y=2.2727,
                stddev_x=3.3166,
                stddev_y=1.6787,
                slope=0.0,
                intercept=2.2727,
                correlation=0.0,
                centroid=5.0,
                sigma=2.0,
                min_x=1,
                max_x=10,
                min_y=0,
                max_y=5,
                x_at_max_y=5,
            ),
            4,
        ],
        [
            dict(y=[0, 1, 0]),
            dict(
                n=3,
                centroid=1,
                sigma=0,
                max_y=1,
                min_y=0,
                min_x=1,
                max_x=2,
                x_at_max_y=1,
                mean_x=1,
                mean_y=1 / 3,
                stddev_x=1,
                stddev_y=0.58,
                correlation=0,
                intercept=0.33,
                slope=0,
            ),
            2,
        ],
    ],
)
def test_analyze_1D(data, expected, ndigits):
    x_arr = data.get("x")
    y_arr = data.get("y")
    results = analyze_1D(y_arr, x_arr)
    assert isinstance(results, dict)

    for k in expected.keys():
        # fmt: off
        assert (
            round(results[k], ndigits) == round(expected[k], ndigits)
        ), f"{k=} {results=} {expected=}"
        # fmt: on


@pytest.mark.parametrize(
    "data, expected, ndigits",
    [
        [
            dict(
                y=[
                    [0, 1, 2, 1, 0],
                    [1, 2, 3, 2, 1],
                    [2, 3, 4, 10, 2],
                    [1, 2, 3, 2, 1],
                ]
            ),
            dict(
                n=(5, 4),
                centroid=(2.1628, 1.8140),
                sigma=(1.1192, 0.8695),
                peak_position=(3, 2),
                max_y=10,
            ),
            4,
        ],
        [
            dict(
                y=[
                    [0, 100, 0],
                    [0, 0, 0],
                ]
            ),
            dict(
                n=(3, 2),
                centroid=(1, 0),
                sigma=(0, 0),
                peak_position=(1, 0),
                max_y=100,
            ),
            1,
        ],
    ],
)
def test_analyze_2D(data, expected, ndigits):
    image = data.get("y")
    results = analyze_2D(image)
    assert isinstance(results, dict)

    def compare_tuples(a, b, label):
        assert len(a) == len(b), label
        for r, e in zip(a, b):
            assert round(r, ndigits) == round(e, ndigits), label

    for k in expected.keys():
        assert k in results, k
        if isinstance(results[k], tuple):
            compare_tuples(results[k], expected[k], f"{k=} {results=}")
        else:
            assert round(results[k], ndigits) == round(expected[k], ndigits), results



"""
test override_parameters module
"""

from ...utils import OverrideParameters
import numpy as np
import pandas as pd
import pytest


@pytest.fixture(scope="function")
def overrides():
    return OverrideParameters()


def test_initial(overrides):
    assert overrides is not None
    assert len(overrides._parameters) == 0


def test_missing(overrides):
    parm = "no such parameter"
    assert parm not in overrides._parameters
    assert overrides.pick(parm, "test value") == "test value"


def test_register(overrides):
    parm = "test_parameter"
    assert parm not in overrides._parameters
    assert len(overrides._parameters) == 0

    overrides.register(parm)
    assert parm in overrides._parameters
    assert len(overrides._parameters) == 1

    overrides.register(parm)
    assert len(overrides._parameters) == 1


def test_undefined(overrides):
    parm = "test_parameter"
    overrides.register(parm)
    assert overrides._parameters[parm] == overrides.undefined
    assert overrides.pick(parm, overrides.undefined) == overrides.undefined
    assert overrides.pick(parm, "undefined") != overrides.undefined
    assert overrides.pick(parm, "undefined") == "undefined"

    overrides.set(parm, 1)
    assert overrides.pick(parm, "undefined") != overrides.undefined
    assert overrides.pick(parm, "undefined") == 1


def test_reset(overrides):
    parm = "test_parameter"
    overrides.register(parm)
    assert overrides.pick(parm, "undefined") == "undefined"

    overrides.set(parm, 1)
    assert overrides.pick(parm, "undefined") == 1

    overrides.reset(parm)
    assert overrides.pick(parm, "undefined") == "undefined"


def test_reset_all(overrides):
    parm = "test_parameter"
    overrides.register(parm)
    assert overrides.pick(parm, "undefined") == "undefined"

    overrides.set(parm, 1)
    assert overrides.pick(parm, "undefined") == 1

    overrides.reset_all()
    assert overrides.pick(parm, "undefined") == "undefined"


def test_summary_type(overrides):
    assert isinstance(overrides.summary(), pd.DataFrame)


def test_summary(overrides):
    parm = "test_parameter"
    assert len(overrides.summary()) == 0

    overrides.register(parm)
    summary = overrides.summary()
    assert len(summary) != 0

    assert summary.columns[0] == "parameter"
    assert summary.columns[1] == "value"

    expected = np.array([[parm, overrides.undefined]])
    assert not (expected == summary.values).all()

    expected = np.array([[parm, "--undefined--"]])
    assert (expected == summary.values).all()



"""
APS User Proposal and ESAF Information
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~ApsBssUserInfoDevice
"""

from ophyd import Component
from ophyd import Device
from ophyd import EpicsSignal


class ApsBssUserInfoDevice(Device):
    """
    Provide current experiment info from the APS BSS.

    .. index:: Ophyd Device; ApsBssUserInfoDevice

    BSS: Beamtime Scheduling System

    EXAMPLE::

        bss_user_info = ApsBssUserInfoDevice(
                            "9id_bss:",
                            name="bss_user_info")
        sd.baseline.append(bss_user_info)

    NOTE: There is info provided by the APS proposal & ESAF systems.
    """

    proposal_number = Component(EpicsSignal, "proposal_number")
    activity = Component(EpicsSignal, "activity", string=True)
    badge = Component(EpicsSignal, "badge", string=True)
    bss_name = Component(EpicsSignal, "bss_name", string=True)
    contact = Component(EpicsSignal, "contact", string=True)
    email = Component(EpicsSignal, "email", string=True)
    institution = Component(EpicsSignal, "institution", string=True)
    station = Component(EpicsSignal, "station", string=True)
    team_others = Component(EpicsSignal, "team_others", string=True)
    time_begin = Component(EpicsSignal, "time_begin", string=True)
    time_end = Component(EpicsSignal, "time_end", string=True)
    timestamp = Component(EpicsSignal, "timestamp", string=True)
    title = Component(EpicsSignal, "title", string=True)
    # not yet updated, see: https://git.aps.anl.gov/jemian/aps_bss_user_info/issues/10
    esaf = Component(EpicsSignal, "esaf", string=True)
    esaf_contact = Component(EpicsSignal, "esaf_contact", string=True)
    esaf_team = Component(EpicsSignal, "esaf_team", string=True)


# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



from ..area_detector_support import AD_EpicsHdf5FileName
from ..area_detector_support import AD_plugin_primed
from ..area_detector_support import AD_prime_plugin2
from bluesky import plans as bp
from bluesky import RunEngine
from ophyd.areadetector import ADComponent, SimDetectorCam
from ophyd.areadetector import SimDetector, SingleTrigger
from ophyd.areadetector.filestore_mixins import FileStoreIterativeWrite
from ophyd.areadetector.plugins import HDF5Plugin_V34 as HDF5Plugin
from ophyd.areadetector.plugins import ImagePlugin_V34 as ImagePlugin
import databroker
import h5py
import pathlib


IOC = "ad:"
IMAGE_DIR = "images"
AD_IOC_MOUNT_PATH = pathlib.Path("/tmp")
BLUESKY_MOUNT_PATH = pathlib.Path(
    f"/tmp/docker_ioc/ioc{IOC.rstrip(':')}{AD_IOC_MOUNT_PATH}"
)

# MUST end with a `/`, pathlib will NOT provide it
WRITE_PATH_TEMPLATE = f"{AD_IOC_MOUNT_PATH / IMAGE_DIR}/"
READ_PATH_TEMPLATE = f"{BLUESKY_MOUNT_PATH / IMAGE_DIR}/"


class myHdf5EpicsIterativeWriter(AD_EpicsHdf5FileName, FileStoreIterativeWrite):
    """Build a custom iterative writer using the override."""

    pass


class myHDF5FileNames(HDF5Plugin, myHdf5EpicsIterativeWriter):
    """Make a custom HDF5 plugin using the override."""

    pass


# define support for the detector (ADSimDetector here):
class MySimDetector(SingleTrigger, SimDetector):
    """SimDetector with HDF5 file names specified by EPICS."""

    _default_read_attrs = ["hdf1"]

    cam = ADComponent(SimDetectorCam, "cam1:")
    image = ADComponent(ImagePlugin, "image1:")

    hdf1 = ADComponent(
        myHDF5FileNames,
        suffix="HDF1:",
        write_path_template=WRITE_PATH_TEMPLATE,
        read_path_template=READ_PATH_TEMPLATE,
    )


def test_no_unexpected_key_in_datum_kwarg():
    """
    Test that this exception is fixed:

    TypeError: __call__() got an unexpected keyword argument 'HDF5_file_name'
    """
    cat = databroker.temp().v2
    RE = RunEngine({})
    RE.subscribe(cat.v1.insert)

    camera = MySimDetector(IOC, name="camera")
    assert isinstance(camera, MySimDetector)

    if hasattr(camera.hdf1.stage_sigs, "array_counter"):
        # remove this so array counter is not set to zero each staging
        del camera.hdf1.stage_sigs["array_counter"]

    camera.cam.stage_sigs["num_images"] = 1
    camera.hdf1.stage_sigs["num_capture"] = 1
    camera.hdf1.stage_sigs["file_template"] = "%s%s_%3.3d.h5"
    camera.hdf1.stage_sigs["compression"] = "zlib"

    # ensure the capture is always staged last
    camera.hdf1.stage_sigs["capture"] = camera.hdf1.stage_sigs.pop("capture")

    camera.wait_for_connection()
    assert camera.connected

    if not AD_plugin_primed(camera.hdf1):
        # camera.hdf1.warmup()
        AD_prime_plugin2(camera.hdf1)

    file_name = "test_image"
    file_path = AD_IOC_MOUNT_PATH / IMAGE_DIR
    file_template = "%s%s_%4.4d.h5"

    camera.hdf1.create_directory.put(-5)
    camera.hdf1.file_name.put(file_name)
    camera.hdf1.file_path.put(str(file_path))
    camera.hdf1.stage_sigs["file_template"] = file_template

    file_name = camera.hdf1.file_name.get()

    camera.stage()
    _fn, _rp, _wp = camera.hdf1.make_filename()
    assert _fn == file_name
    assert _rp == str(BLUESKY_MOUNT_PATH / IMAGE_DIR) + "/"
    assert _wp == str(AD_IOC_MOUNT_PATH / IMAGE_DIR) + "/"
    assert _rp == READ_PATH_TEMPLATE
    assert _wp == WRITE_PATH_TEMPLATE
    assert _rp == camera.hdf1.read_path_template
    assert _wp == camera.hdf1.write_path_template + "/"
    camera.unstage()
    # assert False, "diagnostic stop"

    # take the image
    file_number = camera.hdf1.file_number.get()
    uids = RE(bp.count([camera], num=1))
    assert len(uids) == 1

    uid = uids[0]
    assert isinstance(uid, str)

    # Can bluesky see the new image file?
    ioc_file_name = (file_template % (file_path, "/" + file_name, file_number))
    assert camera.hdf1.full_file_name.get() == ioc_file_name

    assert str(BLUESKY_MOUNT_PATH).startswith("/tmp/docker_ioc/")
    assert BLUESKY_MOUNT_PATH.exists(), BLUESKY_MOUNT_PATH

    # get the full file name relative to the bluesky file directory
    _n = len(camera.hdf1._root.as_posix())
    resource_path = ioc_file_name[_n:].lstrip("/")
    image_file = BLUESKY_MOUNT_PATH.parent / resource_path
    assert image_file.exists(), (
        f"BLUESKY_MOUNT_PATH={BLUESKY_MOUNT_PATH}"
        f" resource_path={resource_path}"
        f" image_file={image_file}"
    )

    # verify that the problem key in datum_kwargs is not found
    for uid in uids:
        uid_start = None
        uid_resource = None
        id_datum = None
        for doc_type, doc in cat.v1[uid].documents():
            if doc_type == "start":
                uid_start = doc["uid"]

            elif doc_type == "datum":
                id_datum = f"{uid_resource}/0"
                assert doc["datum_id"] == id_datum, doc
                # "HDF5_file_name" is an unexpected key in datum_kwargs
                assert doc["datum_kwargs"] == {"point_number": 0}, doc
                assert doc["resource"] == uid_resource, doc

            elif doc_type == "resource":
                uid_resource = doc["uid"]
                assert doc["path_semantics"] == "posix", doc
                assert doc["resource_kwargs"] == {"frame_per_point": 1}, doc
                # full file name is found by databroker at f"{root}/{resource_path}"
                # root: relocatable part where databroker catalog is installed now
                # resource_path: persistent path to file
                assert doc["root"] == camera.hdf1._root.as_posix(), doc
                assert doc["resource_path"].endswith(resource_path)
                _r = str(image_file).lstrip(doc["root"]).lstrip("/")
                assert doc["resource_path"] == _r
                assert doc["run_start"] == uid_start, doc
                assert doc["spec"] == "AD_HDF5", doc

    run = cat.v2[uid]
    assert run is not None

    primary = run.primary
    rsrc = primary._resources
    assert isinstance(rsrc, list), rsrc
    assert len(rsrc) == 1
    rsrc = rsrc[0].to_dict()
    assert isinstance(rsrc, dict), rsrc
    assert rsrc["path_semantics"] == "posix", rsrc
    assert rsrc["resource_kwargs"] == {"frame_per_point": 1}, rsrc
    _r = str(image_file).lstrip(rsrc["root"]).lstrip("/")
    assert rsrc["resource_path"] == _r, rsrc
    assert rsrc["root"] == camera.hdf1._root.as_posix(), rsrc
    assert rsrc["run_start"] == uid_start, rsrc
    assert rsrc["spec"] == "AD_HDF5", rsrc
    assert rsrc["uid"] == uid_resource, rsrc

    # Could databroker find this file?
    image_file = BLUESKY_MOUNT_PATH.parent / resource_path
    # How WE can find it.
    assert image_file.exists()
    # How databroker finds it.
    assert (pathlib.Path(rsrc["root"]) / rsrc["resource_path"]).exists(), rsrc

    # Could databroker read this image?
    with h5py.File(str(image_file), "r") as h5:
        key = "/entry/data/data"
        assert key in h5
        data = h5[key]
        assert data is not None
        assert data.shape == (1, 1024, 1024)

    # If this errors, one cause is issue 651:
    ds = run.primary.read()
    assert ds is not None
    assert camera.image.name in ds



"""
Ophyd support for the EPICS busy record


Public Structures

.. autosummary::

    ~BusyRecord

"""

from ophyd.device import Device, Component
from ophyd import EpicsSignal


class BusyRecord(Device):
    """
    EPICS synApps busy record

    .. index:: Ophyd Device; synApps BusyRecord
    """

    state = Component(EpicsSignal, "", kind="hinted")
    output_link = Component(EpicsSignal, ".OUT", kind="config")
    forward_link = Component(EpicsSignal, ".FLNK", kind="config")

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
Support for IPython profiles
+++++++++++++++++++++++++++++++++++++++

.. autosummary::

   ~getDefaultNamespace
   ~ipython_profile_name
   ~ipython_shell_namespace
"""

import logging


logger = logging.getLogger(__name__)


def getDefaultNamespace(attr="user_ns"):
    """
    get the IPython shell's namespace dictionary (or globals() if not found)
    """
    try:
        from IPython import get_ipython

        ns = getattr(get_ipython(), attr)
    except (ModuleNotFoundError, AttributeError):
        ns = globals()
    return ns


def ipython_profile_name():
    """
    return the name of the current ipython profile or ``None``

    Example (add to default RunEngine metadata)::

        RE.md['ipython_profile'] = str(ipython_profile_name())
        print("using profile: " + RE.md['ipython_profile'])

    """
    from IPython import get_ipython

    return get_ipython().profile


def ipython_shell_namespace():
    """
    get the IPython shell's namespace dictionary (or empty if not found)
    """
    try:
        from IPython import get_ipython

        ns = get_ipython().user_ns
    except AttributeError:
        ns = {}
    return ns

# -----------------------------------------------------------------------------
# :author:    Pete R. Jemian
# :email:     jemian@anl.gov
# :copyright: (c) 2017-2022, UChicago Argonne, LLC
#
# Distributed under the terms of the Argonne National Laboratory Open Source License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# -----------------------------------------------------------------------------



"""
test issue listplans() command
"""

from bluesky import Msg
from bluesky import plans as bp
import pandas as pd
import pytest

from ...devices import SimulatedApsPssShutterWithStatus
from .. import listplans


@pytest.fixture(scope="function")
def ns():
    try:
        from IPython import get_ipython

        ns = get_ipython().user_global_ns
    except (ModuleNotFoundError, AttributeError):
        ns = globals()
    return ns


def my_generator_function():
    yield 1


class MyClass:
    def a_generator_function_method(self):
        yield 1

    def regular_method(self):
        pass

    def undecorated_plan(self):
        yield Msg("testing")


my_class = MyClass()
shutter = SimulatedApsPssShutterWithStatus(name="shutter")


def test_basic():
    result = listplans()
    assert result is not None
    assert isinstance(result, pd.DataFrame)
    assert len(result) == 0


@pytest.mark.parametrize(
    "item, exists",
    [
        ["test_globals", True],
        ["my_generator_function", True],
        ["MyClass", True],
        ["my_class", True],
        ["my_class.a_generator_function_method", False],
        ["my_class.regular_method", False],
        ["hollerin_down_the_hall", False],
    ],
)
def test_globals(item, exists, ns):
    assert (item in ns) == exists, f"item={item}  exists={exists}"


@pytest.mark.parametrize(
    "item, exists",
    [
        ["test_globals", False],
        ["my_generator_function", False],
        ["MyClass", False],
        ["my_class", False],
        ["my_class.a_generator_function_method", False],
        ["my_class.regular_method", False],
        ["a_generator_function_method", True],
        ["regular_method", False],
        ["undecorated_plan", True],
        ["hollerin_down_the_hall", False],
    ],
)
def test_in_class(item, exists):
    result = listplans(my_class)
    generators = result["plan"].to_list()
    assert (item in generators) == exists, f"item={item}  exists={exists} generators={generators}"


@pytest.mark.parametrize(
    "item, plan_name",
    [
        [0, "shutter.get_instantiated_signals"],
        [2, "shutter.walk_signals"],
        [4, "shutter.walk_subdevices"],
    ],
)
def test_shutter(item, plan_name):
    result = listplans(shutter)
    assert result["plan"][item] == plan_name


@pytest.mark.parametrize(
    "item, plan_name",
    [
        [0, "bluesky.plans.adaptive_scan"],
        [4, "bluesky.plans.inner_product_scan"],
        [34, "bluesky.plans.x2x_scan"],
    ],
)
def test_bluesky_plans(item, plan_name):
    result = listplans(bp)
    assert result["plan"][item] == plan_name


@pytest.mark.parametrize(
    "library, length",
    [
        [None, 0],
        [bp, 35],
        [pytest, 0],
        [shutter, 5],
        [globals(), 0],
        [my_class, 2],
        ["ns", 0],
    ],
)
def test_number(library, length, ns):
    if library == "ns":
        library = ns
    result = listplans(library)
    assert len(result) == length
